# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/handlers/tepco.ipynb.

# %% auto 0
__all__ = ['renaming_rules_rdn', 'renaming_rules_cols', 'coi', 'kw', 'load_data', 'FixMissingValuesCB', 'RemoveJapanaseCharCB',
           'FixRangeValueStringCB', 'RemapRdnNameCB', 'time_parser', 'ParseTimeCB', 'RenameColumnCB', 'SelectColumnsCB',
           'get_attrs']

# %% ../../nbs/handlers/tepco.ipynb 7
def load_data(fname_in):
    "Load TEPCO seawater data"
    
    # get measurements
    df_meas = pd.read_excel(fname_in, 
                            skiprows=1, nrows=23643,
                            converters={'Sampling time': lambda x: x if x != '' else '00:00:00'})
    # get locations
    df_loc = pd.read_excel(fname_in, skiprows=23647, nrows=48, usecols=[0,1,2])
    # join
    df = pd.merge(df_meas, df_loc, how="inner", on='Sampling point number')
    # df.drop(columns=['Sampling point number'], inplace=True)
    df.index.name = 'sample'
    return {'seawater': df}

# %% ../../nbs/handlers/tepco.ipynb 15
class FixMissingValuesCB(Callback):
    "Assign `NaN` to values equal to `ND` (not detected) - to be confirmed "
    def __call__(self, tfm): 
        for k in tfm.dfs.keys():
            predicate = tfm.dfs[k] == 'ND'
            tfm.dfs[k][predicate] = np.nan

# %% ../../nbs/handlers/tepco.ipynb 18
class RemoveJapanaseCharCB(Callback):
    "Remove 約 (about) char"
    def _transform_if_about(self, value, about_char='約'):
        if pd.isna(value): return value
        return (value.replace(about_char, '') if str(value).count(about_char) != 0 
                else value)
    
    def __call__(self, tfm): 
        for k in tfm.dfs.keys():
            cols_rdn = [c for c in tfm.dfs[k].columns if ('(Bq/L)' in c) and (tfm.dfs[k][c].dtype == 'object')]
            tfm.dfs[k][cols_rdn] = tfm.dfs[k][cols_rdn].map(self._transform_if_about)

# %% ../../nbs/handlers/tepco.ipynb 21
class FixRangeValueStringCB(Callback):
    "Replace e.g `4.0E+00<&<8.0E+00` by its mean (here 6)"
    
    def _extract_and_calculate_mean(self, s):
        float_strings = re.findall(r"[+-]?\d+\.\d+E[+-]\d+", s)
        float_numbers = np.array(float_strings, dtype=float)
        return float_numbers.mean()

    def _transform_if_range(self, value, range_pattern=r'<&<'):
        if pd.isna(value): return value
        pattern = re.compile(range_pattern)
        return (self._extract_and_calculate_mean(value) if pattern.search(str(value)) 
                else value)

    def __call__(self, tfm): 
        for k in tfm.dfs.keys():
            cols_rdn = [c for c in tfm.dfs[k].columns if ('(Bq/L)' in c) and (tfm.dfs[k][c].dtype == 'object')]
            tfm.dfs[k][cols_rdn] = tfm.dfs[k][cols_rdn].map(self._transform_if_range)

# %% ../../nbs/handlers/tepco.ipynb 24
# Define nuclides-related columns renaming rules
renaming_rules_rdn = {
    '131I radioactivity concentration (Bq/L)': 'i131',
    '131I detection limit (Bq/L)': 'i131_dl',
    '134Cs radioactivity concentration (Bq/L)': 'cs134',
    '134Cs detection limit (Bq/L)': 'cs134_dl',
    '137Cs radioactivity concentration (Bq/L)': 'cs137',
    '137Cs detection limit (Bq/L)': 'cs137_dl',
    '132I radioactivity concentration (Bq/L)': 'i132',
    '132I detection limit (Bq/L)': 'i132_dl',
    '132Te radioactivity concentration (Bq/L)': 'te132',
    '132Te detection limit (Bq/L)': 'te132_dl',
    '136Cs radioactivity concentration (Bq/L)': 'cs136',
    '136Cs detection limit (Bq/L)': 'cs136_dl',
    '140La radioactivity concentration (Bq/L)': 'la140',
    '140La detection limit (Bq/L)': 'la140_dl',
    '89Sr radioactivity concentration (Bq/L)': 'sr89',
    '89Sr detection limit (Bq/L)': 'sr89_dl',
    '90Sr radioactivity concentration (Bq/L)': 'sr90',
    '90Sr detection limit (Bq/L)': 'sr90_dl',
    '238Pu radioactivity concentration (Bq/L)': 'pu238',
    '238Pu detection limit (Bq/L)': 'pu238_dl',
    '239Pu+240Pu radioactivity concentration (Bq/L)': 'pu239_240_tot',
    '239Pu+240Pu statistical error (Bq/L)': 'pu239_240_tot_unc',
    '239Pu+240Pu detection limit (Bq/L)': 'pu239_240_tot_dl',
    'Total alpha radioactivity concentration (Bq/L)': 'talpha',
    'Total alpha detection limit (Bq/L)': 'talpha_dl',
    'Total beta radioactivity concentration (Bq/L)': 'tbeta',
    'Total beta detection limit (Bq/L)': 'tbeta_dl',
    '140Ba radioactivity concentration (Bq/L)': 'ba140',
    '140Ba detection limit (Bq/L)': 'ba140_dl',
    '106Ru radioactivity concentration (Bq/L)': 'ru106',
    '106Ru detection limit (Bq/L)': 'ru106_dl',
    '58Co radioactivity concentration (Bq/L)': 'co58',
    '58Co detection limit (Bq/L)': 'co58_dl',
    '60Co radioactivity concentration (Bq/L)': 'co60',
    '60Co detection limit (Bq/L)': 'co60_dl',
    '144Ce radioactivity concentration (Bq/L)': 'ce144',
    '144Ce detection limit (Bq/L)': 'ce144_dl',
    '54Mn radioactivity concentration (Bq/L)': 'mn54',
    '54Mn detection limit (Bq/L)': 'mn54_dl',
    '3H radioactivity concentration (Bq/L)': 'h3',
    '3H detection limit (Bq/L)': 'h3_dl',
    '125Sb radioactivity concentration (Bq/L)': 'sb125',
    '125Sb detection limit (Bq/L)': 'sb125_dl',
    '105Ru radioactivity concentration (Bq/L)': 'ru105',
    '105Ru detection limit (Bq/L)': 'ru105_dl'}


# %% ../../nbs/handlers/tepco.ipynb 26
class RemapRdnNameCB(Callback):
    "Remap to MARIS radionuclide names"
    def __init__(self,
                 renaming_rules=renaming_rules_rdn):
        fc.store_attr()

    def __call__(self, tfm):
        for k in tfm.dfs.keys():
            tfm.dfs[k].rename(columns=self.renaming_rules, inplace=True)


# %% ../../nbs/handlers/tepco.ipynb 30
def time_parser(col):
    day = str(col.iloc[0].date())
    time = str(col.iloc[1])
    return datetime.strptime(day + ' ' + time, '%Y-%m-%d %H:%M:%S')

# %% ../../nbs/handlers/tepco.ipynb 31
class ParseTimeCB(Callback):
    def __init__(self, 
                 fn_parser=time_parser,
                 cols_time=['Sampling date', 'Sampling time']):
        fc.store_attr()
        
    def __call__(self, tfm):
        for k in tfm.dfs.keys():
            tfm.dfs[k]['time'] = tfm.dfs[k][self.cols_time].apply(self.fn_parser, axis=1)
            tfm.dfs[k].drop(columns=self.cols_time)

# %% ../../nbs/handlers/tepco.ipynb 34
renaming_rules_cols = {
    'Sampling coordinate North latitude (Decimal)': 'lat',
    'Sampling coordinate East longitude (Decimal)': 'lon'}

# %% ../../nbs/handlers/tepco.ipynb 35
class RenameColumnCB(Callback):
    "Normalizing, renaming columns"
    def __init__(self,
                 renaming_rules=renaming_rules_cols):
        fc.store_attr()

    def __call__(self, tfm):
        for k in tfm.dfs.keys():
            tfm.dfs[k].rename(columns=self.renaming_rules, inplace=True)


# %% ../../nbs/handlers/tepco.ipynb 38
coi = ['time', 'lat', 'lon'] + list(renaming_rules_rdn.values())

# %% ../../nbs/handlers/tepco.ipynb 39
class SelectColumnsCB(Callback):
    def __init__(self, coi=coi): fc.store_attr()

    def __call__(self, tfm):
        for k in tfm.dfs.keys():
            tfm.dfs[k] = tfm.dfs[k][self.coi]


# %% ../../nbs/handlers/tepco.ipynb 52
kw = ['oceanography', 'Earth Science > Oceans > Ocean Chemistry> Radionuclides',
      'Earth Science > Human Dimensions > Environmental Impacts > Nuclear Radiation Exposure',
      'Earth Science > Oceans > Water Quality > Ocean Contaminants']


# %% ../../nbs/handlers/tepco.ipynb 53
def get_attrs(tfm, zotero_key, kw=kw):
    return GlobAttrsFeeder(tfm.dfs, cbs=[BboxCB(),
                                    # DepthRangeCB(),
                                    TimeRangeCB(cfg()),
                                    ZoteroCB(zotero_key, cfg=cfg()),
                                    KeyValuePairCB('keywords', ', '.join(kw)),
                                    KeyValuePairCB('publisher_postprocess_logs', ', '.join(tfm.logs))])()

# %% ../../nbs/handlers/tepco.ipynb 56
# def units_fn(grp_name): 
#     return 'Bq/l'

# %% ../../nbs/handlers/tepco.ipynb 57
# def encode(fname_in, fname_out, nc_tpl_path):
#     dfs = load_data(fname_in)
#     tfm = Transformer(dfs, cbs=[FixMissingValuesCB(),
#                                 RemoveJapanaseCharCB(),
#                                 FixRangeValueStringCB(),
#                                 RemapRdnNameCB(),
#                                 ParseTimeCB(),
#                                 RenameColumnCB(),
#                                 SelectColumnsCB(),
#                                 EncodeTimeCB(),
#                                 SanitizeLonLatCB()])
    
#     dfs_tfm = tfm()
#     attrs = get_attrs(tfm, zotero_key='26VMZZ2Q', kw=kw)
#     to_netcdf(dfs_tfm, nc_tpl_path, fname_out, attrs, units_fn)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/cli/init.ipynb.

# %% auto 0
__all__ = ['main']

# %% ../../nbs/cli/init.ipynb 2
from pathlib import Path
import sys

from fastcore.script import *
from fastcore.xtras import mkdir
from fastcore.style import S

from ..configs import base_path, CONFIGS, CONFIGS_CDL, CFG_FNAME, CDL_FNAME
from ..inout import write_toml
from ..utils import download_files_in_folder
from . import create_nc_template

# %% ../../nbs/cli/init.ipynb 3
@call_parse
def main():
    "Create configuration files & download lookup tables"
    path = base_path()
    if path.exists():
        msg = S.red('Configuration files already exist. Do you want to overwrite them? (y/[n]): ')
        if not input(msg).lower().startswith('y'): 
            sys.exit(1)

    mkdir(path, overwrite=True)
    for k, v in CONFIGS['dirs'].items():
        mkdir(v, parents=True, overwrite=True)
            
    # Write configs dictionary as `.toml` in base path
    # using config dictionary literals during initialization
    write_toml(path / CFG_FNAME, CONFIGS)
    write_toml(path / CDL_FNAME, CONFIGS_CDL) 
    
    # Download luts (look-up table) to base path
    print('Downloading look-up tables ...')
    owner, repo = CONFIGS['gh'].values()
    
    download_files_in_folder(owner, repo,
                             src_dir=CONFIGS['paths']['luts'], 
                             dest_dir=CONFIGS['dirs']['lut'])
    
    create_nc_template.main()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/cli/create_nc_template.ipynb.

# %% auto 0
__all__ = ['main']

# %% ../../nbs/cli/create_nc_template.ipynb 1
from fastcore.script import *

from ..nc_template import NCTemplater
from ..configs import (base_path, cdl_cfg, lut_path, 
                             nuc_lut_path, nc_tpl_name, get_enum_dicts)

# %% ../../nbs/cli/create_nc_template.ipynb 2
@call_parse
def main(verbose:bool=False # Verbose
         ):
    "Create MARIS NetCDF template, optionally in verbose mode"
    print('Creating MARIS NetCDF template ...')
    cdl = cdl_cfg()
    templater = NCTemplater(cdl=cdl,
                            nuclide_vars_fname=nuc_lut_path(), 
                            tpl_fname=base_path() / nc_tpl_name(),
                            enum_dicts=get_enum_dicts(lut_src_dir=lut_path(), 
                                                      cdl_enums=cdl['enums']),
                            verbose=verbose
                            )

    templater.generate()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/cli/netcdfy.ipynb.

# %% auto 0
__all__ = ['import_handler', 'main']

# %% ../../nbs/cli/netcdfy.ipynb 2
from fastcore.script import *
import importlib

from ..configs import nc_tpl_path

# %% ../../nbs/cli/netcdfy.ipynb 3
def import_handler(handler_name, fn_name='encode'):
    try:
        handler = importlib.import_module(handler_name)
        return getattr(handler, fn_name)

    except (ImportError, AttributeError):
        print(f"Failed to import function: {fn_name}")

# %% ../../nbs/cli/netcdfy.ipynb 4
@call_parse
def main(handler_name: str,  # Handler's name (e.g helcom, ...)
         src: str,           # Path to dataset to encode
         dest: str,          # Path to converted NetCDF4
         ):
    "Encode MARIS dataset as NetCDF"
    print(f'Encoding: {handler_name} ...')
    encode = import_handler(f'marisco.handlers.{handler_name}')
    encode(src, dest, get_nc_tpl_path())



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/inout.ipynb.

# %% auto 0
__all__ = ['write_toml', 'read_toml']

# %% ../nbs/api/inout.ipynb 2
import tomli_w
import tomli

# %% ../nbs/api/inout.ipynb 3
def write_toml(fname, cfg):
    print(f'Creating {fname}')
    with open(fname, "wb") as f:
        tomli_w.dump(cfg, f)

# %% ../nbs/api/inout.ipynb 4
def read_toml(fname):
    with open(fname, "rb") as f:
        config = tomli.load(f)
    return config



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/nc_template.ipynb.

# %% auto 0
__all__ = ['NCTemplater']

# %% ../nbs/api/nc_template.ipynb 3
from typing import Dict, Union, Callable
from copy import deepcopy
import re

import netCDF4
from netCDF4 import Dataset
import numpy as np
import pandas as pd
from pathlib import Path
import fastcore.all as fc
from fastcore.basics import patch

from .inout import read_toml
from .configs import name2grp, get_enum_dicts

# %% ../nbs/api/nc_template.ipynb 5
class NCTemplater:
    "MARIS NetCDF template generator."
    def __init__(self, 
                 cdl:Dict, # "Pseudo CDL" (`.toml`)
                 nuclide_vars_fname:str, # File name and path of MARIS nuclide lookup table containing variable names
                 tpl_fname:str, # File name and path of NetCDF4 file to be generated
                 enum_dicts:Dict, # MARIS NetCDF enumeration types
                 verbose=False
                ):
        fc.store_attr()
        self.dim = cdl['dim']
        self.enum_types = {}

# %% ../nbs/api/nc_template.ipynb 9
@patch
def nuclide_vars(
    self:NCTemplater,
    col_varnames:str='nc_name', # Column name in the Excel lookup file containing the NetCDF variable names
    col_stdnames:str='nusymbol', # Column name Excel lookup file containing the NetCDF standard names
    dtype:str='f4', # Default data type
    ) -> list[dict]: # List of nuclide variables (including their names and attributes)
    "Return the name of the radionuclide variables analysed."
    df = pd.read_excel(self.nuclide_vars_fname, index_col=0)
    df = df[df.nuclide != 'NOT AVAILABLE']
    return [
        {
            'name': n,
            'dtype': dtype,
            'attrs': {
                'long_name': f"{nuclide.capitalize()} {massnb}",
                'standard_name': sn,
            }
        }
        for n, nuclide, massnb, sn in zip(
            df[col_varnames],
            df['nuclide'].str.capitalize(),
            df['massnb'].astype(str),
            df[col_stdnames],
        )
    ]

# %% ../nbs/api/nc_template.ipynb 12
@patch
def derive(
    self:NCTemplater,
    nuclide:dict, # Nuclide variable name and associated netcdf attributes
    suffix:dict,  # Naming rules as described in CDL (e.g `_unc`)
) -> dict: # Derived variable name and associated attributes
    "Derive NetCDF nuclide-dependent variable names & attributes as defined in CDL." 
    return {
        'name': nuclide['name'] + '_' + suffix['name'],
        'dtype': suffix['dtype'],  # Using dtype from suffix
        'attrs': {key: nuclide['attrs'][key] + suffix['attrs'][key] for key in nuclide['attrs']}
        }

# %% ../nbs/api/nc_template.ipynb 18
@patch
def create_enum_types(self:NCTemplater):
    "Create enumeration types"
    for name, enum in self.enum_dicts.items(): 
        if self.verbose: print(f'Creating {name} enumeration type')
        self.enum_types[name] = self.nc.createEnumType(np.uint, name, enum)

# %% ../nbs/api/nc_template.ipynb 19
@patch
def create_groups(self:NCTemplater):
    "Create NetCDF groups"
    grp_names = [v['name'] for k, v in self.cdl['grps'].items()]
    for grp_name in grp_names:
        grp = self.nc.createGroup(grp_name)
        self.create_variables(grp)

# %% ../nbs/api/nc_template.ipynb 20
@patch
def create_variables(self:NCTemplater, 
                     grp:netCDF4.Group, # NetCDF group
                     ):
        "Create variables"
        self.create_variable(grp, self.dim) # Dimension variable
        self.create_default_variables(grp)
        self.create_group_specific_variables(grp)
        self.create_analyte_variables(grp)

# %% ../nbs/api/nc_template.ipynb 21
@patch
def create_default_variables(self:NCTemplater, 
                             grp:netCDF4.Group, # NetCDF group
                             ):
        "Create Default variables"
        vars = self.cdl['vars']['defaults'].values()
        for var in vars: self.create_variable(grp, var)

# %% ../nbs/api/nc_template.ipynb 22
@patch
def create_group_specific_variables(self:NCTemplater, 
                             grp:netCDF4.Group, # NetCDF group
                             ):
        "Create group specific variables"
        vars = self.cdl['vars']
        for var in vars.get(name2grp(grp.name, self.cdl), {}).values(): 
            self.create_variable(grp, var)

# %% ../nbs/api/nc_template.ipynb 23
@patch
def create_analyte_variables(self:NCTemplater, 
                             grp:netCDF4.Group, # NetCDF group
                             ):
    "Create analyte variables and dependent one as uncertainty, detection limit, ..."    
    for var in self.nuclide_vars():
        self.create_variable(grp, var)
        for v in self.cdl['vars']['suffixes'].values(): 
            self.create_variable(grp, self.derive(var, v))

# %% ../nbs/api/nc_template.ipynb 24
@patch
def create_variable(self:NCTemplater, 
                    grp:netCDF4.Group, # NetCDF group
                    var:Dict, # Variable specificiation dict with `name`, `dtype` and `attrs` keys
                    ):
    "Create NetCDF variable with proper types (standard and enums)"
    name, dtype, attrs = var.values()
    nc_var = grp.createVariable(name, 
                                self.enum_types.get(dtype) or dtype, 
                                self.dim['name'])
    nc_var.setncatts(attrs) 

# %% ../nbs/api/nc_template.ipynb 26
@patch
def generate(self:NCTemplater):
    "Generate CDL"
    # with NetCDFWriter(self.tpl_fname) as self.nc:
    with Dataset(self.tpl_fname, 'w', format='NETCDF4') as self.nc:
        self.nc.setncatts(self.cdl['global_attrs']) 
        self.create_enum_types()
        self.nc.createDimension(self.dim['name'], None) 
        self.create_groups()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/handlers/maris_dump.ipynb.

# %% auto 0
__all__ = ['kw', 'load_dump', 'load_data', 'get_zotero_key', 'get_fname', 'get_varnames_lut', 'RemapRdnNameCB', 'renaming_rules',
           'RenameColumnCB', 'DropNAColumnsCB', 'get_dl_lut', 'SanitizeDetectionLimitCB', 'ParseTimeCB',
           'ReshapeLongToWide', 'get_attrs', 'enums_xtra', 'encode']

# %% ../../nbs/handlers/maris_dump.ipynb 4
from tqdm import tqdm
from pathlib import Path
import fastcore.all as fc
import pandas as pd

from ..callbacks import (Callback, Transformer, SanitizeLonLatCB, EncodeTimeCB)
from ..metadata import (GlobAttrsFeeder, BboxCB,
                              DepthRangeCB, TimeRangeCB,
                              ZoteroCB, KeyValuePairCB)
from ..configs import lut_path, cdl_cfg, cfg, nc_tpl_path, Enums
from ..serializers import NetCDFEncoder

# %% ../../nbs/handlers/maris_dump.ipynb 9
def load_dump(fname): 
    return pd.read_csv(fname, sep='\t', encoding='ISO-8859-1')   

def load_data(df:pd.DataFrame, # MARIS global dump 
                 ref_id:int, # Reference id of interest
                 ):
    "Load specific MARIS dataset through its ref_id"
    lut = {
        'Sediment': 'sediment',
        'Seawater': 'seawater',
        'Suspended matter': 'suspended-matter',
        'Biota': 'biota'}
    dfs = {}
    for name, grp in df[df.ref_id  == ref_id].groupby('samptype'): 
        dfs[lut[name]] = grp
    return dfs

def get_zotero_key(dfs):
    return dfs[next(iter(dfs))][['zoterourl']].iloc[0].values[0].split('/')[-1]

def get_fname(dfs):
    id, name = dfs[next(iter(dfs))][['ref_id', 'displaytext']].iloc[0]
    name = name.replace(',', '').replace('.', '').replace('-', ' ').split(' ')
    return '-'.join(([str(id)] + name)) + '.nc'

# %% ../../nbs/handlers/maris_dump.ipynb 18
def get_varnames_lut():
    fname = lut_path() / 'dbo_nuclide.xlsx'
    df_nuclide = pd.read_excel(fname, usecols=['nuclide_id', 'nc_name'])
    return df_nuclide.set_index('nuclide_id').to_dict()['nc_name']

# %% ../../nbs/handlers/maris_dump.ipynb 19
class RemapRdnNameCB(Callback):
    "Remap to MARIS radionuclide names."
    def __init__(self,
                 fn_lut=get_varnames_lut):
        fc.store_attr()

    def __call__(self, tfm):
        lut = self.fn_lut()
        for k in tfm.dfs.keys():
            tfm.dfs[k]['nuclide_id'] = tfm.dfs[k]['nuclide_id'].replace(lut)

# %% ../../nbs/handlers/maris_dump.ipynb 24
# To be added: endperiod, totdepth, lab
def renaming_rules():
    vars = cdl_cfg()['vars']
    # Define column names renaming rules
    return {
        'latitude': vars['defaults']['lat']['name'],
        'longitude': vars['defaults']['lon']['name'],
        'begperiod': vars['defaults']['time']['name'],
        'sampdepth': vars['defaults']['depth']['name'],
        'uncertaint': vars['suffixes']['uncertainty']['name'],
        'unit_id': vars['suffixes']['unit']['name'],
        'detection': vars['suffixes']['detection_limit']['name'],
        'area_id': vars['defaults']['area']['name'], 
        'species_id': vars['bio']['species']['name'],
        'biogroup_id': vars['bio']['bio_group']['name'],
        'bodypar_id': vars['bio']['body_part']['name'],
        'sedtype_id': vars['sed']['sed_type']['name'],
        'volume': vars['suffixes']['volume']['name'],
        'salinity': vars['suffixes']['salinity']['name'],
        'temperatur': vars['suffixes']['temperature']['name'],
        'sampmet_id': vars['suffixes']['sampling_method']['name'],
        'prepmet_id': vars['suffixes']['preparation_method']['name'],
        'counmet_id': vars['suffixes']['counting_method']['name'],
        'activity': 'value',
        'nuclide_id': 'nuclide'
    }

# %% ../../nbs/handlers/maris_dump.ipynb 25
class RenameColumnCB(Callback):
    "Renaming variables to MARIS standard names."
    def __init__(self, renaming_rules=renaming_rules): fc.store_attr()
    def __call__(self, tfm):
        lut = renaming_rules()
        coi = lut.keys()
        for k in tfm.dfs.keys():
            # Select cols of interest
            tfm.dfs[k] = tfm.dfs[k].loc[:, coi]
            # Rename cols
            tfm.dfs[k].rename(columns=lut, inplace=True)

# %% ../../nbs/handlers/maris_dump.ipynb 28
class DropNAColumnsCB(Callback):
    "Drop variable containing only NaN or 'Not available' (id=0 in MARIS lookup tables)."
    def __init__(self, na_value=0):
        fc.store_attr()
        
    def isMarisNA(self, col): 
        return len(col.unique()) == 1 and col.iloc[0] == self.na_value
    
    def dropMarisNA(self, df):
        na_cols = [col for col in df.columns if self.isMarisNA(df[col])]
        return df.drop(labels=na_cols, axis=1)
        
    def __call__(self, tfm):
        for k in tfm.dfs.keys():
            tfm.dfs[k] = tfm.dfs[k].dropna(axis=1, how='all')
            tfm.dfs[k] = self.dropMarisNA(tfm.dfs[k])

# %% ../../nbs/handlers/maris_dump.ipynb 31
def get_dl_lut():
    fname = lut_path() / 'dbo_detectlimit.xlsx'
    df_nuclide = pd.read_excel(fname, usecols=['name', 'id'])
    return df_nuclide.set_index('name').to_dict()['id']

# %% ../../nbs/handlers/maris_dump.ipynb 33
class SanitizeDetectionLimitCB(Callback):
    "Assign Detection Limit name to its id based on MARIS nomenclature."
    def __init__(self,
                 fn_lut=get_dl_lut):
        fc.store_attr()

    def __call__(self, tfm):
        lut = self.fn_lut()
        for k in tfm.dfs.keys():
            tfm.dfs[k]['dl'] = tfm.dfs[k]['dl'].replace(lut)

# %% ../../nbs/handlers/maris_dump.ipynb 36
class ParseTimeCB(Callback):
    def __call__(self, tfm):
        for k in tfm.dfs.keys():
            tfm.dfs[k]['time'] = pd.to_datetime(tfm.dfs[k].time, format='ISO8601')

# %% ../../nbs/handlers/maris_dump.ipynb 40
class ReshapeLongToWide(Callback):
    "Convert data from long to wide with renamed columns."
    def __init__(self, columns='nuclide', values=['value']):
        fc.store_attr()
        # Retrieve all possible derived vars (e.g 'unc', 'dl', ...) from configs
        self.derived_cols = [value['name'] for value in cdl_cfg()['vars']['suffixes'].values()]
    
    def renamed_cols(self, cols):
        "Flatten columns name"
        return [inner if outer == "value" else f'{inner}_{outer}'
                if inner else outer
                for outer, inner in cols]

    def pivot(self, df):
        # Among all possible 'derived cols' select the ones present in df
        derived_coi = [col for col in self.derived_cols if col in df.columns]
        
        df.reset_index(names='sample', inplace=True)
        
        idx = list(set(df.columns) - set([self.columns] + derived_coi + self.values))
        # print('idx: ', idx)
        # print('derived_coi: ', derived_coi)
        return df.pivot_table(index=idx,
                              columns=self.columns,
                              values=self.values + derived_coi,
                              fill_value=0,
                              aggfunc=lambda x: x
                              ).reset_index()

    def __call__(self, tfm):
        for k in tfm.dfs.keys():
            tfm.dfs[k] = self.pivot(tfm.dfs[k])
            tfm.dfs[k].columns = self.renamed_cols(tfm.dfs[k].columns)
            # tfm.dfs[k].index.name = 'sample'
            tfm.dfs[k].set_index('sample', inplace=True)

# %% ../../nbs/handlers/maris_dump.ipynb 49
kw = ['oceanography', 'Earth Science > Oceans > Ocean Chemistry> Radionuclides',
      'Earth Science > Human Dimensions > Environmental Impacts > Nuclear Radiation Exposure',
      'Earth Science > Oceans > Ocean Chemistry > Ocean Tracers, Earth Science > Oceans > Marine Sediments',
      'Earth Science > Oceans > Ocean Chemistry, Earth Science > Oceans > Sea Ice > Isotopes',
      'Earth Science > Oceans > Water Quality > Ocean Contaminants',
      'Earth Science > Biological Classification > Animals/Vertebrates > Fish',
      'Earth Science > Biosphere > Ecosystems > Marine Ecosystems',
      'Earth Science > Biological Classification > Animals/Invertebrates > Mollusks',
      'Earth Science > Biological Classification > Animals/Invertebrates > Arthropods > Crustaceans',
      'Earth Science > Biological Classification > Plants > Macroalgae (Seaweeds)']

# %% ../../nbs/handlers/maris_dump.ipynb 50
def get_attrs(tfm, zotero_key='26VMZZ2Q', kw=kw):
    return GlobAttrsFeeder(tfm.dfs, cbs=[
        BboxCB(),
        DepthRangeCB(),
        TimeRangeCB(cfg()),
        ZoteroCB(zotero_key, cfg=cfg()),
        KeyValuePairCB('keywords', ', '.join(kw)),
        KeyValuePairCB('publisher_postprocess_logs', ', '.join(tfm.logs))
        ])()

# %% ../../nbs/handlers/maris_dump.ipynb 52
def enums_xtra(tfm, vars):
    "Retrieve a subset of the lengthy enum as 'species_t' for instance"
    enums = Enums(lut_src_dir=lut_path(), cdl_enums=cdl_cfg()['enums'])
    xtras = {}
    for var in vars:
        unique_vals = tfm.unique(var)
        if unique_vals.any():
            xtras[f'{var}_t'] = enums.filter(f'{var}_t', unique_vals)
    return xtras

# %% ../../nbs/handlers/maris_dump.ipynb 53
def encode(fname_in, fname_out, nc_tpl_path, **kwargs):
    df = load_dump(fname_in)
    ref_ids = kwargs.get('ref_ids', df.ref_id.unique())
    print('Encoding ...')
    for ref_id in tqdm(ref_ids, leave=False):
        dfs = load_data(df, ref_id)
        print(get_fname(dfs))
        tfm = Transformer(dfs, cbs=[
            RemapRdnNameCB(),
            RenameColumnCB(),
            DropNAColumnsCB(),
            SanitizeDetectionLimitCB(),
            ParseTimeCB(),
            ReshapeLongToWide(),
            EncodeTimeCB(cfg()),
            SanitizeLonLatCB(verbose=True)
            ])
       
        tfm()
        encoder = NetCDFEncoder(tfm.dfs, 
                                src_fname=nc_tpl_path,
                                dest_fname=Path(fname_out) / get_fname(dfs), 
                                global_attrs=get_attrs(tfm, zotero_key=get_zotero_key(dfs), kw=kw),
                                verbose=kwargs.get('verbose', False),
                                enums_xtra=enums_xtra(tfm, vars=['species', 'body_part'])
                                )
        encoder.encode()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/handlers/helcom.ipynb.

# %% auto 0
__all__ = ['varnames_lut_updates', 'coi_units_unc', 'coi_grp', 'renaming_rules', 'kw', 'load_data', 'rename_cols',
           'LowerStripRdnNameCB', 'get_unique_nuclides', 'get_varnames_lut', 'RemapRdnNameCB', 'ParseTimeCB',
           'fix_units', 'NormalizeUncUnitCB', 'get_maris_species', 'get_worms_species', 'LookupBiotaSpeciesCB',
           'get_bodypart', 'LookupBiotaBodyPartCB', 'get_sediment', 'LookupSedimentCB', 'RenameColumnCB',
           'ReshapeLongToWide', 'get_attrs', 'encode']

# %% ../../nbs/handlers/helcom.ipynb 4
import pandas as pd
from tqdm import tqdm
from functools import partial
import fastcore.all as fc

from pathlib import Path

from ..utils import (has_valid_varname, match_worms, 
                           match_maris_species, match_maris_sediment)
from ..callbacks import (Callback, Transformer,
                               EncodeTimeCB, SanitizeLonLatCB)

from ..metadata import (GlobAttrsFeeder, BboxCB,
                              DepthRangeCB, TimeRangeCB,
                              ZoteroCB, KeyValuePairCB)

from ..configs import base_path, nc_tpl_path, cfg, cache_path
from ..serializers import NetCDFEncoder

# %% ../../nbs/handlers/helcom.ipynb 8
def load_data(src_dir,
                smp_types=['SEA', 'SED', 'BIO']):
    "Load HELCOM data and return them as individual dataframe by sample type"
    dfs = {}
    lut_smp_type = {'SEA': 'seawater', 'SED': 'sediment', 'BIO': 'biota'}
    for smp_type in smp_types:
        fname_meas = smp_type + '02.csv'
        fname_smp = smp_type + '01.csv'
        df = pd.merge(pd.read_csv(Path(src_dir)/fname_meas),  # measurements
                      pd.read_csv(Path(src_dir)/fname_smp),  # sample
                      on='KEY', how='left')
        dfs[lut_smp_type[smp_type]] = df
    return dfs


def rename_cols(cols):
    "Flatten multiindex columns"
    new_cols = []
    for outer, inner in cols:
        if not inner:
            new_cols.append(outer)
        else:
            if outer == 'unc':
                new_cols.append(inner + '_' + outer)
            if outer == 'value':
                new_cols.append(inner)
    return new_cols

# %% ../../nbs/handlers/helcom.ipynb 17
class LowerStripRdnNameCB(Callback):
    "Convert nuclide names to lowercase & strip any trailing space(s)"

    def __call__(self, tfm):
        for k in tfm.dfs.keys():
            tfm.dfs[k]['NUCLIDE'] = tfm.dfs[k]['NUCLIDE'].apply(
                lambda x: x.lower().strip())

# %% ../../nbs/handlers/helcom.ipynb 20
def get_unique_nuclides(dfs):
    "Get list of unique radionuclide types measured across samples."
    nuclides = []
    for k in dfs.keys():
        nuclides += dfs[k]['NUCLIDE'].unique().tolist()
    return nuclides

# %% ../../nbs/handlers/helcom.ipynb 22
varnames_lut_updates = {
    'k-40': 'k40',
    'cm243244': 'cm243_244_tot',
    'cs134137': 'cs134_137_tot',
    'pu239240': 'pu239_240_tot',
    'pu238240': 'pu238_240_tot'}


# %% ../../nbs/handlers/helcom.ipynb 23
def get_varnames_lut(dfs, lut=varnames_lut_updates):
    lut = {n: n for n in set(get_unique_nuclides(dfs))}
    lut.update(varnames_lut_updates)
    return lut


# %% ../../nbs/handlers/helcom.ipynb 25
class RemapRdnNameCB(Callback):
    "Remap to MARIS radionuclide names."
    def __init__(self,
                 fn_lut=partial(get_varnames_lut, lut=varnames_lut_updates)):
        fc.store_attr()

    def __call__(self, tfm):
        lut = self.fn_lut(tfm.dfs)
        for k in tfm.dfs.keys():
            tfm.dfs[k]['NUCLIDE'].replace(lut, inplace=True)

# %% ../../nbs/handlers/helcom.ipynb 29
class ParseTimeCB(Callback):
    def __call__(self, tfm):
        for k in tfm.dfs.keys():
            tfm.dfs[k]['time'] = pd.to_datetime(tfm.dfs[k].DATE, 
                                                format='%m/%d/%y %H:%M:%S')

# %% ../../nbs/handlers/helcom.ipynb 32
# Make measurement and uncertainty units consistent
def fix_units(df, meas_col, unc_col):
    return df.apply(lambda row: row[unc_col] * row[meas_col]/100, axis=1)

# %% ../../nbs/handlers/helcom.ipynb 33
# Columns of interest
coi_units_unc = [('seawater', 'VALUE_Bq/m³', 'ERROR%_m³'),
                 ('biota', 'VALUE_Bq/kg', 'ERROR%'),
                 ('sediment', 'VALUE_Bq/kg', 'ERROR%_kg')]

# %% ../../nbs/handlers/helcom.ipynb 34
class NormalizeUncUnitCB(Callback):
    "Convert uncertainty from % to activity unit"

    def __init__(self, coi=coi_units_unc): fc.store_attr()

    def __call__(self, tfm):
        for grp, val, unc in self.coi:
            tfm.dfs[grp][unc] = self.fix_units(tfm.dfs[grp], val, unc)

    def fix_units(self, df, meas_col, unc_col):
        return df.apply(lambda row: row[unc_col] * row[meas_col]/100, axis=1)

# %% ../../nbs/handlers/helcom.ipynb 35
class NormalizeUncUnitCB(Callback):
    "Convert uncertainty from % to activity unit"

    def __init__(self, coi=coi_units_unc): fc.store_attr()

    def __call__(self, tfm):
        for grp, val, unc in self.coi:
            tfm.dfs[grp][unc] = self.fix_units(tfm.dfs[grp], val, unc)

    def fix_units(self, df, meas_col, unc_col):
        return df.apply(lambda row: row[unc_col] * row[meas_col]/100, axis=1)

# %% ../../nbs/handlers/helcom.ipynb 41
def get_maris_species(fname_in, fname_cache, overwrite=False, verbose=False):
    fname_cache = cache_path() / fname_cache
    lut = {}
    df = pd.read_csv(Path(fname_in) / 'RUBIN_NAME.csv')
    
    if overwrite or (not fname_cache.exists()):
        if verbose: print('Source:Destination')    
        for _, row in tqdm(df.iterrows(), total=df.shape[0]):
            match = match_maris_species(row['SCIENTIFIC NAME'])
            lut[row['RUBIN']] = {'id': match.iloc[0]['species_id'], 'name': match.iloc[0]['species']}
            if verbose: print(f'{row["SCIENTIFIC NAME"]}: {match.iloc[0]["species"]}')
        fc.save_pickle(fname_cache, lut)
    else:
        lut = fc.load_pickle(fname_cache)
        
    return lut        

# %% ../../nbs/handlers/helcom.ipynb 43
def get_worms_species(fname_in, fname_cache, overwrite=False):
    fname_cache = cache_path() / fname_cache
    lut = {}

    if overwrite or (not fname_cache.exists()):
        df = pd.read_csv(Path(fname_in) / 'RUBIN_NAME.csv')
        
        for _, row in tqdm(df[['RUBIN', 'SCIENTIFIC NAME']].iterrows(), total=df.shape[0]):
            res = match_worms(row['SCIENTIFIC NAME'])
            if (res == -1):
                print(f"No match for {row['RUBIN']} ({row['SCIENTIFIC NAME']})")
                id = -1
                lut[row['RUBIN']] = {'id': id, 'name': '', 'status': '', 'match_type': ''}
            else:
                if len(res[0]) > 1:
                    print(f"Several matches for {row['RUBIN']} ({row['SCIENTIFIC NAME']})")
                    
                id, name, status, match_type = [res[0][0].get(key) 
                                                for key in ['AphiaID', 'scientificname', 'status', 'match_type']]        
                
                lut[row['RUBIN']] = {'id': id, 'name': name, 'status': status, 'match_type': match_type}
        fc.save_pickle(fname_cache, lut)
    else:
        lut = fc.load_pickle(fname_cache)
        
    return lut

# %% ../../nbs/handlers/helcom.ipynb 45
class LookupBiotaSpeciesCB(Callback):
    'Match species with MARIS database.'
    def __init__(self, fn_lut): fc.store_attr()
    def __call__(self, tfm):
        lut = self.fn_lut()
        tfm.dfs['biota']['species_id'] = tfm.dfs['biota']['RUBIN'].apply(
            lambda x: lut[x.strip()]['id'])

# %% ../../nbs/handlers/helcom.ipynb 49
def get_bodypart():
    "Naive lut - TO BE REFACTORED"
    return {
        5: 52, 1: 1,
        41: 1, 3: 3,
        51: 54, 43: 19,
        42: 59, 12: 20,
        10: 7, 18: 25,
        52: 55, 20: 38,
        8: 12, 54: 57,
        53: 56}

# %% ../../nbs/handlers/helcom.ipynb 50
class LookupBiotaBodyPartCB(Callback):
    'Update bodypart id based on MARIS dbo_bodypar.xlsx'
    def __init__(self, fn_lut): fc.store_attr()
    def __call__(self, tfm):
        lut = self.fn_lut()
        tfm.dfs['biota']['body_part'] = tfm.dfs['biota']['TISSUE'].apply(lambda x: lut[x])

# %% ../../nbs/handlers/helcom.ipynb 55
def get_sediment(verbose=False):
    lut = {}
    if verbose: print('Source:Destination')
    df_sediment = pd.read_csv(Path(fname_in) / 'SEDIMENT_TYPE.csv')
    for _, row in df_sediment.iterrows():
        match = match_maris_sediment(row['SEDIMENT TYPE'])
        lut[row['SEDI']] = match.iloc[0,0]
        if verbose: print(f'({row["SEDI"]}) {row["SEDIMENT TYPE"]}: ({match.iloc[0,0]}) {match.iloc[0,1]}')
    return lut        

# %% ../../nbs/handlers/helcom.ipynb 60
class LookupSedimentCB(Callback):
    'Update sediment id  based on MARIS dbo_sedtype.xlsx'
    def __init__(self, fn_lut): fc.store_attr()
    def __call__(self, tfm):
        lut = self.fn_lut()
        tfm.dfs['sediment']['SEDI'] = dfs['sediment']['SEDI'].fillna(-99).astype('int')
        # To check with Helcom
        tfm.dfs['sediment']['SEDI'].replace(56, -99, inplace=True)
        tfm.dfs['sediment']['SEDI'].replace(73, -99, inplace=True)
        tfm.dfs['sediment']['sed_type'] = tfm.dfs['sediment']['SEDI'].apply(lambda x: lut[x])

# %% ../../nbs/handlers/helcom.ipynb 63
# Define columns of interest by sample type
coi_grp = {'seawater': ['NUCLIDE', 'VALUE_Bq/m³', 'ERROR%_m³', 'time',
                        'TDEPTH', 'LATITUDE (dddddd)', 'LONGITUDE (dddddd)'],
           'sediment': ['NUCLIDE', 'VALUE_Bq/kg', 'ERROR%_kg', 'time',
                        'TDEPTH', 'LATITUDE (dddddd)', 'LONGITUDE (dddddd)',
                        'sed_type'],
           'biota': ['NUCLIDE', 'VALUE_Bq/kg', 'ERROR%', 'time',
                     'SDEPTH', 'LATITUDE ddmmmm', 'LONGITUDE ddmmmm',
                     'species_id', 'body_part']}


# %% ../../nbs/handlers/helcom.ipynb 64
# Define column names renaming rules
renaming_rules = {
    'NUCLIDE': 'nuclide',
    'VALUE_Bq/m³': 'value',
    'VALUE_Bq/kg': 'value',
    'ERROR%_m³': 'unc',
    'ERROR%_kg': 'unc',
    'ERROR%': 'unc',
    'TDEPTH': 'depth',
    'SDEPTH': 'depth',
    'LATITUDE (dddddd)': 'lat',
    'LATITUDE ddmmmm': 'lat',
    'LONGITUDE (dddddd)': 'lon',
    'LONGITUDE ddmmmm': 'lon'
}


# %% ../../nbs/handlers/helcom.ipynb 65
class RenameColumnCB(Callback):
    def __init__(self,
                 coi=coi_grp,
                 renaming_rules=renaming_rules):
        fc.store_attr()

    def __call__(self, tfm):
        for k in tfm.dfs.keys():
            # Select cols of interest
            tfm.dfs[k] = tfm.dfs[k].loc[:, self.coi[k]]

            # Rename cols
            tfm.dfs[k].rename(columns=self.renaming_rules, inplace=True)

# %% ../../nbs/handlers/helcom.ipynb 68
class ReshapeLongToWide(Callback):
    def __init__(self): fc.store_attr()

    def __call__(self, tfm):
        for k in tfm.dfs.keys():
            cols = ['nuclide']
            vals = ['value', 'unc']
            idx = list(set(tfm.dfs[k].columns) -
                       set(cols + vals))  # All others

            tfm.dfs[k] = tfm.dfs[k].pivot_table(index=idx,
                                                columns=cols,
                                                values=vals).reset_index()

            # Flatten cols name
            tfm.dfs[k].columns = rename_cols(tfm.dfs[k].columns)

            # Set index
            tfm.dfs[k].index.name = 'sample'

# %% ../../nbs/handlers/helcom.ipynb 78
kw = ['oceanography', 'Earth Science > Oceans > Ocean Chemistry> Radionuclides',
      'Earth Science > Human Dimensions > Environmental Impacts > Nuclear Radiation Exposure',
      'Earth Science > Oceans > Ocean Chemistry > Ocean Tracers, Earth Science > Oceans > Marine Sediments',
      'Earth Science > Oceans > Ocean Chemistry, Earth Science > Oceans > Sea Ice > Isotopes',
      'Earth Science > Oceans > Water Quality > Ocean Contaminants',
      'Earth Science > Biological Classification > Animals/Vertebrates > Fish',
      'Earth Science > Biosphere > Ecosystems > Marine Ecosystems',
      'Earth Science > Biological Classification > Animals/Invertebrates > Mollusks',
      'Earth Science > Biological Classification > Animals/Invertebrates > Arthropods > Crustaceans',
      'Earth Science > Biological Classification > Plants > Macroalgae (Seaweeds)']


# %% ../../nbs/handlers/helcom.ipynb 79
def get_attrs(tfm, zotero_key='26VMZZ2Q', kw=kw):
    return GlobAttrsFeeder(tfm.dfs, cbs=[
        BboxCB(),
        DepthRangeCB(),
        TimeRangeCB(cfg()),
        ZoteroCB(zotero_key, cfg=cfg()),
        KeyValuePairCB('keywords', ', '.join(kw)),
        KeyValuePairCB('publisher_postprocess_logs', ', '.join(tfm.logs))
        ])()

# %% ../../nbs/handlers/helcom.ipynb 84
def encode(fname_in, fname_out, nc_tpl_path, **kwargs):
    dfs = load_data(fname_in)
    tfm = Transformer(dfs, cbs=[
        LowerStripRdnNameCB(),
        RemapRdnNameCB(),
        ParseTimeCB(),
        NormalizeUncUnitCB(),
        LookupBiotaSpeciesCB(partial(get_maris_species, 
                                     fname_in, 'species_helcom.pkl')),
        LookupBiotaBodyPartCB(get_bodypart),
        LookupSedimentCB(get_sediment),
        RenameColumnCB(),
        ReshapeLongToWide(),
        EncodeTimeCB(cfg()),
        SanitizeLonLatCB()
        ])
    
    species_lut = get_maris_species(fname_in, 'species_helcom.pkl')
    enums_xtra = {
        'species_t': {info['name']: info['id'] 
                      for info in species_lut.values() if info['name'] != ''}
    }
    
    encoder = NetCDFEncoder(tfm(), 
                            src_fname=nc_tpl_path,
                            dest_fname=fname_out, 
                            global_attrs=get_attrs(tfm, zotero_key='26VMZZ2Q', kw=kw),
                            enums_xtra=enums_xtra,
                            **kwargs)
    encoder.encode()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/metadata.ipynb.

# %% auto 0
__all__ = ['GlobAttrsFeeder', 'BboxCB', 'DepthRangeCB', 'TimeRangeCB', 'ZoteroItem', 'ZoteroCB', 'KeyValuePairCB']

# %% ../nbs/api/metadata.ipynb 2
import pandas as pd
import fastcore.all as fc
from cftime import num2date
from pyzotero import zotero, zotero_errors
import json

from .utils import get_bbox, Callback, run_cbs
from .configs import CONFIGS

# %% ../nbs/api/metadata.ipynb 3
class GlobAttrsFeeder:
    "Produce NetCDF global attributes as specified by the callbacks."
    def __init__(self, 
                 dfs:dict, # Dictionary of NetCDF group DataFrames
                 cbs:list=[], # Callbacks
                 logs:list=[] # List of preprocessing steps taken
                 ): 
        fc.store_attr()
        self.attrs = {}
        
    def callback(self):
        run_cbs(self.cbs, self)
        
    def __call__(self):
        self.callback()
        return self.attrs

# %% ../nbs/api/metadata.ipynb 4
class BboxCB(Callback):
    "Compute dataset geographical bounding box"
    def __call__(self, obj):
        bbox = get_bbox(pd.concat(obj.dfs)) 
        lon_min, lon_max, lat_min, lat_max = [str(bound) for bound in bbox.bounds]
        obj.attrs.update({
            'geospatial_lat_min': lat_min, 
            'geospatial_lat_max': lat_max,
            'geospatial_lon_min': lon_min,
            'geospatial_lon_max': lon_max,
            'geospatial_bounds': bbox.wkt})

# %% ../nbs/api/metadata.ipynb 5
class DepthRangeCB(Callback):
    "Compute depth values range"
    def __init__(self, depth_col='depth'): fc.store_attr()
    def __call__(self, obj):
        depths = pd.concat(obj.dfs).get(self.depth_col, default=pd.Series([]))
        if not depths.empty:
            max_depth, min_depth = depths.max(), depths.min()
            obj.attrs.update({
                'geospatial_vertical_max': '0' if min_depth == 0 else str(-min_depth),
                'geospatial_vertical_min': str(-max_depth)})

# %% ../nbs/api/metadata.ipynb 6
class TimeRangeCB(Callback):
    "Compute time values range"
    def __init__(self, cfg): fc.store_attr()
    def __call__(self, obj):
        time = pd.concat(obj.dfs)['time']
        start, end = [num2date(t, units=self.cfg['units']['time']).isoformat() 
                      for t in (time.min(), time.max())]
        obj.attrs.update({
            'time_coverage_start': start,
            'time_coverage_end': end})

# %% ../nbs/api/metadata.ipynb 7
class ZoteroItem:
    def __init__(self, item_id, cfg):
        self.cfg = cfg
        self.item = self.getItem(item_id)
    
    def exist(self): return self.item != None
    
    def getItem(self, item_id):
        zot = zotero.Zotero(self.cfg['lib_id'], 'group', self.cfg['api_key'])
        try:
            return zot.item(item_id)
        except zotero_errors.ResourceNotFound:
            print(f'Item {item_id} does not exist in Zotero library')
            return None
            
    def title(self):
        return self.item['data']['title']
    
    def summary(self):
        return self.item['data']['abstractNote']
    
    def creator_name(self):
        # creators = [f'{c["creatorType"]}: {c["name"]}' for c in self.item['data']['creators']]
        # return '; '.join(creators)
        return json.dumps(self.item['data']['creators'])
            
    def __repr__(self):
        return json.dumps(self.item, indent=4) 

# %% ../nbs/api/metadata.ipynb 8
class ZoteroCB(Callback):
    "Retrieve Zotero metadata"
    def __init__(self, itemId, cfg): fc.store_attr()
    def __call__(self, obj):
        item = ZoteroItem(self.itemId, self.cfg['zotero'])
        if item.exist(): 
            for attr in ['title', 'summary', 'creator_name']:
                obj.attrs[attr] = getattr(item, attr)()

# %% ../nbs/api/metadata.ipynb 12
class KeyValuePairCB(Callback):
    def __init__(self, k, v): fc.store_attr()
    def __call__(self, obj): obj.attrs[self.k] = self.v



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/callbacks.ipynb.

# %% auto 0
__all__ = ['Callback', 'run_cbs', 'Transformer', 'EncodeTimeCB', 'SanitizeLonLatCB']

# %% ../nbs/api/callbacks.ipynb 2
import copy
import fastcore.all as fc
from operator import attrgetter
from cftime import date2num
import numpy as np

from .configs import cfg

# %% ../nbs/api/callbacks.ipynb 4
class Callback(): order = 0

# %% ../nbs/api/callbacks.ipynb 5
def run_cbs(cbs, obj=None):
    for cb in sorted(cbs, key=attrgetter('order')):
        if cb.__doc__: obj.logs.append(cb.__doc__)
        cb(obj)

# %% ../nbs/api/callbacks.ipynb 6
class Transformer():
    def __init__(self, dfs, cbs=None): 
        self.cbs = cbs
        self.dfs = {k: v.copy() for k, v in dfs.items()}
        self.logs = []
        
    def callback(self):
        run_cbs(self.cbs, self)
        
    def unique(self, col_name):
        "Distinct values of a specific column present in all groups"
        columns = [df.get(col_name) for df in self.dfs.values() if df.get(col_name) is not None]
        values = np.concatenate(columns) if columns else []
        return np.unique(values)
        
    def __call__(self):
        if self.cbs: self.callback()
        return self.dfs

# %% ../nbs/api/callbacks.ipynb 11
class EncodeTimeCB(Callback):
    "Encode time as `int` representing seconds since xxx"    
    def __init__(self, cfg): fc.store_attr()
    def __call__(self, tfm): 
        def format_time(x): return date2num(x, units=self.cfg['units']['time'])
        
        for k in tfm.dfs.keys():
            tfm.dfs[k]['time'] = tfm.dfs[k]['time'].apply(format_time)

# %% ../nbs/api/callbacks.ipynb 12
class SanitizeLonLatCB(Callback):
    "Drop row when both longitude & latitude equal 0."
    def __init__(self, verbose=False): fc.store_attr()
    def __call__(self, tfm):
        for grp, df in tfm.dfs.items():
            mask = (df.lon == 0) & (df.lat == 0)
            nZeroes = mask.sum()
            if nZeroes and self.verbose: 
                print(f'The "{grp}" group contains {nZeroes} data points whose (lon, lat) = (0, 0)')
            tfm.dfs[grp] = df.loc[~mask]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/configs.ipynb.

# %% auto 0
__all__ = ['CFG_FNAME', 'CDL_FNAME', 'NUCLIDE_LOOKUP_FNAME', 'MARISCO_CFG_DIRNAME', 'CONFIGS', 'CONFIGS_CDL',
           'NETCDF_TO_PYTHON_TYPE', 'base_path', 'cfg', 'nuc_lut_path', 'lut_path', 'cache_path', 'cdl_cfg',
           'species_lut_path', 'sediments_lut_path', 'name2grp', 'nc_tpl_name', 'nc_tpl_path', 'sanitize', 'get_lut',
           'Enums', 'get_enum_dicts']

# %% ../nbs/api/configs.ipynb 2
from pathlib import Path
import re
from functools import partial

from .inout import read_toml, write_toml
import pandas as pd

import fastcore.all as fc

# %% ../nbs/api/configs.ipynb 4
CFG_FNAME = 'configs.toml'
CDL_FNAME = 'cdl.toml'
NUCLIDE_LOOKUP_FNAME = 'dbo_nuclide.xlsx'
MARISCO_CFG_DIRNAME = '.marisco'

# %% ../nbs/api/configs.ipynb 5
def base_path(): return Path.home() / MARISCO_CFG_DIRNAME

# %% ../nbs/api/configs.ipynb 8
CONFIGS = {
    'gh': {
        'owner': 'franckalbinet',
        'repo': 'marisco'
    },
    'names': {
        'nc_template': 'maris-template.nc'
    },
    'dirs': {
        'lut': str(base_path() / 'lut'), # Look-up tables
        'cache': str(base_path() / 'cache'), # Cache (e.f WoRMS species)
        'tmp': str(base_path() / 'tmp')
    },
    'paths': {
        'luts': 'nbs/files/lut'
    },
    'units': {
        'time': 'seconds since 1970-01-01 00:00:00.0'
    },
    'zotero': {
        'api_key': '7nuMByrm19RP7reiQtYJwgPk',
        'lib_id': '2432820'
    }
}

# %% ../nbs/api/configs.ipynb 12
def cfg(): return read_toml(base_path() / CFG_FNAME)

# %% ../nbs/api/configs.ipynb 13
def nuc_lut_path(): return Path(cfg()['dirs']['lut']) / NUCLIDE_LOOKUP_FNAME

# %% ../nbs/api/configs.ipynb 14
def lut_path(): return Path(cfg()['dirs']['lut'])

# %% ../nbs/api/configs.ipynb 15
def cache_path(): return Path(cfg()['dirs']['cache'])

# %% ../nbs/api/configs.ipynb 16
CONFIGS_CDL = { 
    'placeholder': '_to_be_filled_in_',
    'grps': {
        'sea': {
            'name': 'seawater'
        },
        'bio': {
            'name': 'biota'
        },
        'sed': {
            'name': 'sediment'
        },
        'sus': {
            'name': 'suspended-matter'
        }
    },
    'global_attrs': {
        # Do not update keys. Only values if required
        'id': '', # zotero?
        'title': '',
        'summary': '',
        'keywords': '',
        'keywords_vocabulary': 'GCMD Science Keywords',
        'keywords_vocabulary_url': 'https://gcmd.earthdata.nasa.gov/static/kms/',
        'record': '',
        'featureType': '',
        'cdm_data_type': '',

        # Conventions
        'Conventions': 'CF-1.10 ACDD-1.3',

        # Publisher [ACDD1.3]
        'publisher_name': 'Paul MCGINNITY, Iolanda OSVATH, Florence DESCROIX-COMANDUCCI',
        'publisher_email': 'p.mc-ginnity@iaea.org, i.osvath@iaea.org, F.Descroix-Comanducci@iaea.org', 
        'publisher_url': 'https://maris.iaea.org',
        'publisher_institution': 'International Atomic Energy Agency - IAEA', 

        # Creator info [ACDD1.3]
        'creator_name': '',
        'institution': '',
        'metadata_link': '',
        'creator_email': '',
        'creator_url': '',
        'references': '',
        'license': ' '.join(['Without prejudice to the applicable Terms and Conditions', 
                             '(https://nucleus.iaea.org/Pages/Others/Disclaimer.aspx),',
                             'I hereby agree that any use of the data will contain appropriate',
                             'acknowledgement of the data source(s) and the IAEA Marine',
                             'Radioactivity Information System (MARIS).']),
        'comment': '',
        # Dataset info & coordinates [ACDD1.3]
        #'project': '', # Network long name
        #'platform': '', # Should be a long / full name
        'geospatial_lat_min': '', 
        'geospatial_lon_min': '',
        'geospatial_lat_max': '',
        'geospatial_lon_max': '',
        'geospatial_vertical_min': '',
        'geospatial_vertical_max': '',
        'geospatial_bounds': '', # wkt representation
        'geospatial_bounds_crs': 'EPSG:4326',

        # Time information
        'time_coverage_start': '',
        'time_coverage_end': '',
        #'time_coverage_resolution': '',
        'local_time_zone': '',
        'date_created': '',
        'date_modified': '',
        #
        # -- Additional metadata (custom to MARIS)
        #
        'publisher_postprocess_logs': ''
        },
    'dim': {
        'name': 'sample',
        'attrs': {
            'long_name': 'Sample ID of measurement'
        },
        'dtype': 'u8'
    },
    'vars': {    
        'defaults': {
            'lon': {
                'name': 'lon',
                'attrs': {
                    'long_name': 'Measurement longitude',
                    'standard_name': 'longitude',
                    'units': 'degrees_north',
                    'axis': 'Y',
                    '_CoordinateAxisType': 'Lon'
                },
                'dtype': 'f4'
            },
            'lat': {
                'name': 'lat',
                'attrs': {
                    'long_name': 'Measurement latitude',
                    'standard_name': 'latitude',
                    'units': 'degrees_east',
                    'axis': 'X',
                    '_CoordinateAxisType': 'Lat'
                },
                'dtype': 'f4'
            },
            'depth': {
                'name': 'depth',
                'attrs': {
                    'long_name': 'Depth below seal level',
                    'standard_name': 'depth_below_sea_floor',
                    'units': 'm',
                    'axis': 'Z'
                },
                'dtype': 'f4'
            },
            'time': {
                'name': 'time',
                'attrs': {
                    'long_name': 'Time of measurement',
                    'standard_name': 'time',
                    'units': 'seconds since 1970-01-01 00:00:00.0',
                    'time_origin': '1970-01-01 00:00:00',
                    'time_zone': 'UTC',
                    'abbreviation': 'Date/Time',
                    'axis': 'T',
                    'calendar': 'gregorian'
                },
                'dtype': 'u8',
            },
            'area': {
                'name': 'area',
                'attrs': {
                    'long_name': 'Marine area/region id',
                    'standard_name': 'area_id'
                },
                'dtype': 'area_t'
            },
        },
        'bio': {
            'bio_group': {
                'name': 'bio_group',
                'attrs': {
                    'long_name': 'Biota group',
                    'standard_name': 'biota_group_tbd'
                },
                'dtype': 'bio_group_t'
            },
            'species': {
                'name': 'species',
                'attrs': {  
                    'long_name': 'Species',
                    'standard_name': 'species'
                },
                'dtype': 'species_t'
            },
            'body_part': {
                'name': 'body_part',
                'attrs': {
                    'long_name': 'Body part',
                    'standard_name': 'body_part_tbd'
                },
                'dtype': 'body_part_t' 
            }
        },
        'sed': {
            'sed_type': {
                'name': 'sed_type',
                'attrs': {
                    'long_name': 'Sediment type',
                    'standard_name': 'sediment_type_tbd'
                },
                'dtype': 'sed_type_t'
            }
        },
        'suffixes':  {
            'uncertainty': {
                'name': 'unc',
                'attrs': {
                    'long_name': ' uncertainty',
                    'standard_name': '_uncertainty'
                },
                'dtype': 'f4'
            },
            'detection_limit': {
                'name': 'dl',
                'attrs': {
                    'long_name': ' detection limit',
                    'standard_name': '_detection_limit'
                },
                'dtype': 'dl_t'
            },
            'volume': {
                'name': 'vol',
                'attrs': {
                    'long_name': ' volume',
                    'standard_name': '_volume'
                },
                'dtype': 'f4'
            },
            'salinity': {
                'name': 'sal',
                'attrs': {
                    'long_name': ' salinity',
                    'standard_name': '_sal'
                },
                'dtype': 'f4'
            },
            'temperature': {
                'name': 'temp',
                'attrs': {
                    'long_name': ' temperature',
                    'standard_name': '_temp'
                },
                'dtype': 'f4'
            },
            'filtered': {
                'name': 'filt',
                'attrs': {
                    'long_name': ' filtered',
                    'standard_name': '_filtered'
                },
                'dtype': 'filt_t'
            },
            'counting_method': {
                'name': 'counmet',
                'attrs': {
                    'long_name': ' counting method',
                    'standard_name': '_counting_method'
                },
                'dtype': 'counmet_t'
            },
            'sampling_method': {
                'name': 'sampmet',
                'attrs': {
                    'long_name': ' sampling method',
                    'standard_name': '_sampling_method'
                },
                'dtype': 'sampmet_t'
            },
            'preparation_method': {
                'name': 'prepmet',
                'attrs': {
                    'long_name': ' preparation method',
                    'standard_name': '_preparation_method'
                },
                'dtype': 'prepmet_t'
            },
            'unit': {
                'name': 'unit',
                'attrs': {
                    'long_name': ' unit',
                    'standard_name': '_unit'
                },
                'dtype': 'unit_t'
            }
        }
    },
    'enums': [
        {
            'name': 'area_t', 
            'fname': 'dbo_area.xlsx', 
            'key': 'displayName', 
            'value':'areaId'
        },
        {
            'name': 'bio_group_t', 
            'fname': 'dbo_biogroup.xlsx', 
            'key': 'biogroup', 
            'value':'biogroup_id'
        },
        {
            'name': 'body_part_t', 
            'fname': 'dbo_bodypar.xlsx', 
            'key': 'bodypar', 
            'value':'bodypar_id'
        },
        {
            'name': 'species_t', 
            'fname': 'dbo_species_cleaned.xlsx', 
            'key': 'species', 
            'value':'species_id'
        },
        {
            'name': 'sed_type_t', 
            'fname': 'dbo_sedtype.xlsx', 
            'key': 'sedtype', 
            'value':'sedtype_id'
        },
        {
            'name': 'unit_t', 
            'fname': 'dbo_unit.xlsx', 
            'key': 'unit_sanitized', 
            'value':'unit_id'
        },
        {
            'name': 'dl_t', 
            'fname': 'dbo_detectlimit.xlsx', 
            'key': 'name_sanitized', 
            'value':'id'
        },
        {
            'name': 'filt_t', 
            'fname': 'dbo_filtered.xlsx', 
            'key': 'name',
            'value':'id'
        },
        {
            'name': 'counmet_t', 
            'fname': 'dbo_counmet.xlsx', 
            'key': 'counmet',
            'value':'counmet_id'
        },
        {
            'name': 'sampmet_t', 
            'fname': 'dbo_sampmet.xlsx', 
            'key': 'sampmet',
            'value':'sampmet_id'
        },
        {
            'name': 'prepmet_t', 
            'fname': 'dbo_prepmet.xlsx', 
            'key': 'prepmet',
            'value':'prepmet_id'
        }
        ]
}

# %% ../nbs/api/configs.ipynb 19
def cdl_cfg(): return read_toml(base_path() / CDL_FNAME)

# %% ../nbs/api/configs.ipynb 20
def species_lut_path():
    src_dir = lut_path()
    fname = [enum for enum in cdl_cfg()['enums'] if enum['name'] == 'species_t'][0]['fname']
    return src_dir / fname

# %% ../nbs/api/configs.ipynb 21
def sediments_lut_path():
    src_dir = lut_path()
    fname = [enum for enum in cdl_cfg()['enums'] if enum['name'] == 'sed_type_t'][0]['fname']
    return src_dir / fname

# %% ../nbs/api/configs.ipynb 23
NETCDF_TO_PYTHON_TYPE = {
    'u8': int,
    'f4': float
    }

# %% ../nbs/api/configs.ipynb 24
def name2grp(
    name:str, # Name of the group
    cdl:dict, # CDL configuration
    ):
    # Reverse `cdl.toml` config group dict so that group config key can be retrieve based on its name
    return {v['name']:k  for k, v in cdl['grps'].items()}[name]

# %% ../nbs/api/configs.ipynb 27
def nc_tpl_name():
    p = base_path()
    return read_toml(p / 'configs.toml')['names']['nc_template']

# %% ../nbs/api/configs.ipynb 28
def nc_tpl_path():
    "Return the name of the MARIS NetCDF template as defined in `configs.toml`"
    p = base_path()
    return p / read_toml(p / 'configs.toml')['names']['nc_template']

# %% ../nbs/api/configs.ipynb 30
def sanitize(s:str # String to sanitize
             ) -> str:
    """
    Sanitize dictionary key to comply with NetCDF enumeration type: 
    
        - remove `(`, `)`, `.`, `/`, `-`  
        - strip the string
    """
    s = re.sub(r'[().]', '', s)
    return re.sub(r'[/-]', ' ', s).strip() 

# %% ../nbs/api/configs.ipynb 34
def get_lut(src_dir:str, # Directory containing lookup tables
            fname:str, # Excel file lookup table name
            key:str, # Excel file column name to be used as dict keys 
            value:str, # Excel file column name to be used as dict values 
            do_sanitize:bool=True # Sanitization required?
            ) -> dict: # MARIS lookup table
    "Convert MARIS db lookup table excel file to dictionary `{'name': id, ...}`."
    fname = Path(src_dir) / fname
    df = pd.read_excel(fname, usecols=[key, value]).dropna(subset=value)
    df[value] = df[value].astype('int')
    df = df.set_index(key)
    lut = df[value].to_dict()
    if do_sanitize: lut = {sanitize(k): v for k, v in lut.items()}
    return lut

# %% ../nbs/api/configs.ipynb 37
class Enums():
    "Return dictionaries of MARIS NetCDF's enumeration types"
    def __init__(self, 
               lut_src_dir:str,
               cdl_enums:dict
               ):
        fc.store_attr()
        self.types = self.lookup()
        
    def filter(self, name, values):
        return {name: id for name, id in self.types[name].items() if id in values}
    
    def lookup(self):
        types = {}
        for enum in self.cdl_enums:
            name, fname, key, value = enum.values()
            lut = get_lut(self.lut_src_dir, fname, key=key, value=value)
            types[name] = lut
        return types

# %% ../nbs/api/configs.ipynb 44
def get_enum_dicts(
    lut_src_dir:str,
    cdl_enums:dict,
    **kwargs
    ):
    "Return a dict of NetCDF enumeration types"
    enum_types = {}
    for enum in cdl_enums:
        name, fname, key, value = enum.values()
        lut = get_lut(lut_src_dir, fname, key=key, value=value, **kwargs)
        enum_types[name] = lut
        
    return enum_types



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/utils.ipynb.

# %% auto 0
__all__ = ['Callback', 'run_cbs', 'Transformer', 'has_valid_varname', 'get_bbox', 'download_files_in_folder', 'download_file',
           'match_worms', 'match_maris_species', 'match_maris_sediment']

# %% ../nbs/api/utils.ipynb 2
from pathlib import Path
from netCDF4 import Dataset
from fastcore.test import test_eq
import fastcore.all as fc
import pandas as pd
import numpy as np
import requests
from shapely import MultiPoint
from operator import attrgetter

from .configs import species_lut_path, sediments_lut_path

import jellyfish as jf
from collections.abc import Callable

# %% ../nbs/api/utils.ipynb 4
class Callback(): order = 0

# %% ../nbs/api/utils.ipynb 5
def run_cbs(cbs, obj=None):
    for cb in sorted(cbs, key=attrgetter('order')):
        if cb.__doc__: obj.logs.append(cb.__doc__)
        cb(obj)

# %% ../nbs/api/utils.ipynb 6
class Transformer():
    def __init__(self, dfs, cbs=None): 
        fc.store_attr()
        self.logs = []
        
    def callback(self):
        run_cbs(self.cbs, self)
        
    def __call__(self):
        self.callback()
        return self.dfs

# %% ../nbs/api/utils.ipynb 8
def has_valid_varname(
    var_names:list, # variable (nuclide) names
    cdl_path:str, # Path to MARIS CDL file (point of truth)
):
    "Check that proposed variable names are in MARIS CDL"
    has_valid = True
    with Dataset(cdl_path) as nc:
        grp = nc.groups[list(nc.groups.keys())[0]] # get any group
        for name in var_names:
            if name not in grp.variables.keys():
                has_valid = False
                print(f'"{name}" variable name not found in MARIS CDL')
    
    return has_valid

# %% ../nbs/api/utils.ipynb 12
def get_bbox(df,
             coord_cols=('lon', 'lat')
            ):
    x, y = coord_cols        
    arr = [(row[x], row[y]) for _, row in df.iterrows()]
    return MultiPoint(arr).envelope

# %% ../nbs/api/utils.ipynb 19
def download_files_in_folder(owner:str, 
                             repo:str, 
                             src_dir:str, 
                             dest_dir:str
                             ):
    "Make a GET request to the GitHub API to get the contents of the folder"
    url = f"https://api.github.com/repos/{owner}/{repo}/contents/{src_dir}"
    response = requests.get(url)

    if response.status_code == 200:
        contents = response.json()

        # Iterate over the files and download them
        for item in contents:
            if item["type"] == "file":
                fname = item["name"]
                download_file(owner, repo, src_dir, dest_dir, fname)
    else:
        print(f"Error: {response.status_code}")

def download_file(owner, repo, src_dir, dest_dir, fname):
    # Make a GET request to the GitHub API to get the raw file contents
    url = f"https://raw.githubusercontent.com/{owner}/{repo}/master/{src_dir}/{fname}"
    response = requests.get(url)

    if response.status_code == 200:
        # Save the file locally
        with open(Path(dest_dir) / fname, "wb") as file:
            file.write(response.content)
        print(f"{fname} downloaded successfully.")
    else:
        print(f"Error: {response.status_code}")

# %% ../nbs/api/utils.ipynb 21
def match_worms(
    name:str # Name of species to look up in WoRMS
    ):
    "Lookup `name` in WoRMS (fuzzy match)"
    url = 'https://www.marinespecies.org/rest/AphiaRecordsByMatchNames'
    params = {
        'scientificnames[]': [name],
        'marine_only': 'true'
    }
    headers = {
        'accept': 'application/json'
    }
    
    response = requests.get(url, params=params, headers=headers)
    
    # Check if the request was successful (status code 200)
    if response.status_code == 200:
        data = response.json()
        return data
    else:
        return -1

# %% ../nbs/api/utils.ipynb 27
def match_maris_species(
    name:str, # Name of species to look up 
    col_lookup:str='species', # Name of the column where the character strings match
    dist_fn:Callable=jf.levenshtein_distance, # Jellyfish distance to use
    coi:list=['species_id', 'species', 'Taxonname', 'TaxonDBID', 
              'AphiaID', 'scientificname', 'status', 'rank', 
              'match_type'], # Columns of interest to display
    nresults:int=10 # Maxiumn number of results to return
    ):
    "Lookup `name` in MARIS master species lookup table"
    df = pd.read_excel(species_lut_path())
    df = df.dropna(subset=col_lookup)
    df = df.astype({'species_id':'int'})
    results = []
    for _, row in df.iterrows():
        score = dist_fn(name.lower(), row[col_lookup].lower())
        result = row[coi].to_dict()
        result['score'] = score
        results.append(result)
    return pd.DataFrame(results).sort_values(by='score', ascending=True)[:nresults]

# %% ../nbs/api/utils.ipynb 30
def match_maris_sediment(
    name:str, # Name of sediment to look up 
    col_lookup:str='sedtype', # Name of the column where the character strings match
    dist_fn:Callable=jf.levenshtein_distance, # Jellyfish distance to use
    coi:list=['sedtype_id', 'sedtype'], # Columns of interest to display
    nresults:int=10 # Maxiumn number of results to return
    ):
    "Lookup `sedtype` in MARIS master sediment lookup table"
    df = pd.read_excel(sediments_lut_path())
    df = df.dropna(subset=col_lookup)
    df = df.astype({'sedtype_id':'int'})
    results = []
    for _, row in df.iterrows():
        score = dist_fn(name.lower(), row[col_lookup].lower())
        result = row[coi].to_dict()
        result['score'] = score
        results.append(result)
    return pd.DataFrame(results).sort_values(by='score', ascending=True)[:nresults]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/serializers.ipynb.

# %% auto 0
__all__ = ['NetCDFEncoder']

# %% ../nbs/api/serializers.ipynb 3
from netCDF4 import Dataset
import pandas as pd
from typing import Dict, Callable
import pandas as pd
from fastcore.basics import patch, store_attr

# %% ../nbs/api/serializers.ipynb 4
class NetCDFEncoder:
    "MARIS NetCDF encoder."
    def __init__(self, 
                 dfs:dict[pd.DataFrame], # dict of Dataframes to encode with group name as key {'sediment': df_sed, ...}
                 src_fname:str, # File name and path to the MARIS CDL template
                 dest_fname:str, # Name of output file to produce
                 global_attrs:Dict, # Global attributes
                 enums_xtra:Dict={}, # Enumeration types to overwrite
                 verbose:bool=False, # Print currently written NetCDF group and variable names
                 ):
        store_attr()
        self.enum_types = {}

# %% ../nbs/api/serializers.ipynb 7
@patch 
def copy_global_attributes(self:NetCDFEncoder):
    "Update NetCDF template global attributes as specified by `global_attrs` argument."
    self.dest.setncatts(self.src.__dict__)
    for k, v in self.global_attrs.items(): self.dest.setncattr(k, v)

# %% ../nbs/api/serializers.ipynb 8
@patch
def copy_dimensions(self:NetCDFEncoder):
    for name, dimension in self.src.dimensions.items():
        self.dest.createDimension(name, (len(dimension) if not dimension.isunlimited() else None))

# %% ../nbs/api/serializers.ipynb 9
@patch
def process_groups(self:NetCDFEncoder):
    for grp_name, df in self.dfs.items():
        self.process_group(grp_name, df)

# %% ../nbs/api/serializers.ipynb 10
@patch
def process_group(self:NetCDFEncoder, group_name, df):
    group_dest = self.dest.createGroup(group_name)
    self.copy_variables(group_name, df, group_dest)

# %% ../nbs/api/serializers.ipynb 11
@patch
def copy_variables(self:NetCDFEncoder, group_name, df, group_dest):
    for var_name, var_src in self.src.groups[group_name].variables.items():
        if var_name in df.reset_index().columns: 
            self.copy_variable(var_name, var_src, df, group_dest)

# %% ../nbs/api/serializers.ipynb 12
@patch
def copy_variable(self:NetCDFEncoder, var_name, var_src, df, group_dest):
    dtype_name = var_src.datatype.name
    enums_src = self.src.enumtypes
    if self.verbose: print(f'Group: {group_dest.name}, Variable: {var_name}')
    # if the type of the var is an enum (meaning present in the template src) then create it
    if dtype_name in enums_src: self.copy_enum_type(dtype_name)   
    self._create_and_copy_variable(var_name, var_src, df, group_dest, dtype_name)
    self.copy_variable_attributes(var_name, var_src, group_dest)

# %% ../nbs/api/serializers.ipynb 13
@patch
def _create_and_copy_variable(self:NetCDFEncoder, var_name, var_src, df, group_dest, dtype_name):
    variable_type = self.enum_types.get(dtype_name, var_src.datatype)
    group_dest.createVariable(var_name, variable_type, var_src.dimensions, compression='zlib', complevel=9)
    
    df_sanitized = self.cast_verbose_rf(df, var_name)
    group_dest[var_name][:] = df_sanitized.values

# %% ../nbs/api/serializers.ipynb 14
@patch
def copy_enum_type(self:NetCDFEncoder, dtype_name):
    # if enum type not already created
    if dtype_name not in self.enum_types:
        enum_info = self.src.enumtypes[dtype_name]
        # if a subset of an enum is defined in enums_xtra (typically for the lengthy species_t)
        if enum_info.name in self.enums_xtra:
            enum_info.enum_dict = self.enums_xtra[enum_info.name]
        self.enum_types[dtype_name] = self.dest.createEnumType(enum_info.dtype, 
                                                               enum_info.name, 
                                                               enum_info.enum_dict)

# %% ../nbs/api/serializers.ipynb 15
@patch
def copy_variable_attributes(self:NetCDFEncoder, var_name, var_src, group_dest):
    group_dest[var_name].setncatts(var_src.__dict__)

# %% ../nbs/api/serializers.ipynb 16
@patch
def cast_verbose_rf(self:NetCDFEncoder, 
                    df, 
                    col):
    """
    Try to cast df column to numeric type:
        - Silently coerce to nan if not possible
        - But log when it failed
    """
    n_before = sum(df.reset_index()[col].notna())
    df_after = pd.to_numeric(df.reset_index()[col],
                                    errors='coerce', downcast=None)
    n_after = sum(df_after.notna())
    if n_before != n_after: 
        print(f'Failed to convert type of {col} in {n_before - n_after} occurences')
    
    return df_after

# %% ../nbs/api/serializers.ipynb 17
@patch
def encode(self:NetCDFEncoder):
    "Encode MARIS NetCDF based on template and dataframes."
    with Dataset(self.src_fname, format='NETCDF4') as self.src, Dataset(self.dest_fname, 'w', format='NETCDF4') as self.dest:
        self.copy_global_attributes()
        self.copy_dimensions()
        self.process_groups()



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/marisco',
                'doc_host': 'https://franckalbinet.github.io',
                'git_url': 'https://github.com/franckalbinet/marisco',
                'lib_path': 'marisco'},
  'syms': { 'marisco.callbacks': { 'marisco.callbacks.Callback': ('api/callbacks.html#callback', 'marisco/callbacks.py'),
                                   'marisco.callbacks.EncodeTimeCB': ('api/callbacks.html#encodetimecb', 'marisco/callbacks.py'),
                                   'marisco.callbacks.EncodeTimeCB.__call__': ('api/callbacks.html#__call__', 'marisco/callbacks.py'),
                                   'marisco.callbacks.EncodeTimeCB.__init__': ('api/callbacks.html#__init__', 'marisco/callbacks.py'),
                                   'marisco.callbacks.SanitizeLonLatCB': ('api/callbacks.html#sanitizelonlatcb', 'marisco/callbacks.py'),
                                   'marisco.callbacks.SanitizeLonLatCB.__call__': ('api/callbacks.html#__call__', 'marisco/callbacks.py'),
                                   'marisco.callbacks.SanitizeLonLatCB.__init__': ('api/callbacks.html#__init__', 'marisco/callbacks.py'),
                                   'marisco.callbacks.Transformer': ('api/callbacks.html#transformer', 'marisco/callbacks.py'),
                                   'marisco.callbacks.Transformer.__call__': ('api/callbacks.html#__call__', 'marisco/callbacks.py'),
                                   'marisco.callbacks.Transformer.__init__': ('api/callbacks.html#__init__', 'marisco/callbacks.py'),
                                   'marisco.callbacks.Transformer.callback': ('api/callbacks.html#callback', 'marisco/callbacks.py'),
                                   'marisco.callbacks.Transformer.unique': ('api/callbacks.html#unique', 'marisco/callbacks.py'),
                                   'marisco.callbacks.run_cbs': ('api/callbacks.html#run_cbs', 'marisco/callbacks.py')},
            'marisco.cli.create_nc_template': { 'marisco.cli.create_nc_template.main': ( 'cli/create_nc_template.html#main',
                                                                                         'marisco/cli/create_nc_template.py')},
            'marisco.cli.init': {'marisco.cli.init.main': ('cli/init.html#main', 'marisco/cli/init.py')},
            'marisco.cli.netcdfy': { 'marisco.cli.netcdfy.import_handler': ('cli/netcdfy.html#import_handler', 'marisco/cli/netcdfy.py'),
                                     'marisco.cli.netcdfy.main': ('cli/netcdfy.html#main', 'marisco/cli/netcdfy.py')},
            'marisco.configs': { 'marisco.configs.Enums': ('api/configs.html#enums', 'marisco/configs.py'),
                                 'marisco.configs.Enums.__init__': ('api/configs.html#__init__', 'marisco/configs.py'),
                                 'marisco.configs.Enums.filter': ('api/configs.html#filter', 'marisco/configs.py'),
                                 'marisco.configs.Enums.lookup': ('api/configs.html#lookup', 'marisco/configs.py'),
                                 'marisco.configs.base_path': ('api/configs.html#base_path', 'marisco/configs.py'),
                                 'marisco.configs.cache_path': ('api/configs.html#cache_path', 'marisco/configs.py'),
                                 'marisco.configs.cdl_cfg': ('api/configs.html#cdl_cfg', 'marisco/configs.py'),
                                 'marisco.configs.cfg': ('api/configs.html#cfg', 'marisco/configs.py'),
                                 'marisco.configs.get_enum_dicts': ('api/configs.html#get_enum_dicts', 'marisco/configs.py'),
                                 'marisco.configs.get_lut': ('api/configs.html#get_lut', 'marisco/configs.py'),
                                 'marisco.configs.lut_path': ('api/configs.html#lut_path', 'marisco/configs.py'),
                                 'marisco.configs.name2grp': ('api/configs.html#name2grp', 'marisco/configs.py'),
                                 'marisco.configs.nc_tpl_name': ('api/configs.html#nc_tpl_name', 'marisco/configs.py'),
                                 'marisco.configs.nc_tpl_path': ('api/configs.html#nc_tpl_path', 'marisco/configs.py'),
                                 'marisco.configs.nuc_lut_path': ('api/configs.html#nuc_lut_path', 'marisco/configs.py'),
                                 'marisco.configs.sanitize': ('api/configs.html#sanitize', 'marisco/configs.py'),
                                 'marisco.configs.sediments_lut_path': ('api/configs.html#sediments_lut_path', 'marisco/configs.py'),
                                 'marisco.configs.species_lut_path': ('api/configs.html#species_lut_path', 'marisco/configs.py')},
            'marisco.handlers.helcom': { 'marisco.handlers.helcom.LookupBiotaBodyPartCB': ( 'handlers/helcom.html#lookupbiotabodypartcb',
                                                                                            'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.LookupBiotaBodyPartCB.__call__': ( 'handlers/helcom.html#__call__',
                                                                                                     'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.LookupBiotaBodyPartCB.__init__': ( 'handlers/helcom.html#__init__',
                                                                                                     'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.LookupBiotaSpeciesCB': ( 'handlers/helcom.html#lookupbiotaspeciescb',
                                                                                           'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.LookupBiotaSpeciesCB.__call__': ( 'handlers/helcom.html#__call__',
                                                                                                    'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.LookupBiotaSpeciesCB.__init__': ( 'handlers/helcom.html#__init__',
                                                                                                    'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.LookupSedimentCB': ( 'handlers/helcom.html#lookupsedimentcb',
                                                                                       'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.LookupSedimentCB.__call__': ( 'handlers/helcom.html#__call__',
                                                                                                'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.LookupSedimentCB.__init__': ( 'handlers/helcom.html#__init__',
                                                                                                'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.LowerStripRdnNameCB': ( 'handlers/helcom.html#lowerstriprdnnamecb',
                                                                                          'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.LowerStripRdnNameCB.__call__': ( 'handlers/helcom.html#__call__',
                                                                                                   'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.NormalizeUncUnitCB': ( 'handlers/helcom.html#normalizeuncunitcb',
                                                                                         'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.NormalizeUncUnitCB.__call__': ( 'handlers/helcom.html#__call__',
                                                                                                  'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.NormalizeUncUnitCB.__init__': ( 'handlers/helcom.html#__init__',
                                                                                                  'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.NormalizeUncUnitCB.fix_units': ( 'handlers/helcom.html#fix_units',
                                                                                                   'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.ParseTimeCB': ( 'handlers/helcom.html#parsetimecb',
                                                                                  'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.ParseTimeCB.__call__': ( 'handlers/helcom.html#__call__',
                                                                                           'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.RemapRdnNameCB': ( 'handlers/helcom.html#remaprdnnamecb',
                                                                                     'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.RemapRdnNameCB.__call__': ( 'handlers/helcom.html#__call__',
                                                                                              'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.RemapRdnNameCB.__init__': ( 'handlers/helcom.html#__init__',
                                                                                              'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.RenameColumnCB': ( 'handlers/helcom.html#renamecolumncb',
                                                                                     'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.RenameColumnCB.__call__': ( 'handlers/helcom.html#__call__',
                                                                                              'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.RenameColumnCB.__init__': ( 'handlers/helcom.html#__init__',
                                                                                              'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.ReshapeLongToWide': ( 'handlers/helcom.html#reshapelongtowide',
                                                                                        'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.ReshapeLongToWide.__call__': ( 'handlers/helcom.html#__call__',
                                                                                                 'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.ReshapeLongToWide.__init__': ( 'handlers/helcom.html#__init__',
                                                                                                 'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.encode': ('handlers/helcom.html#encode', 'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.fix_units': ( 'handlers/helcom.html#fix_units',
                                                                                'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.get_attrs': ( 'handlers/helcom.html#get_attrs',
                                                                                'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.get_bodypart': ( 'handlers/helcom.html#get_bodypart',
                                                                                   'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.get_maris_species': ( 'handlers/helcom.html#get_maris_species',
                                                                                        'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.get_sediment': ( 'handlers/helcom.html#get_sediment',
                                                                                   'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.get_unique_nuclides': ( 'handlers/helcom.html#get_unique_nuclides',
                                                                                          'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.get_varnames_lut': ( 'handlers/helcom.html#get_varnames_lut',
                                                                                       'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.get_worms_species': ( 'handlers/helcom.html#get_worms_species',
                                                                                        'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.load_data': ( 'handlers/helcom.html#load_data',
                                                                                'marisco/handlers/helcom.py'),
                                         'marisco.handlers.helcom.rename_cols': ( 'handlers/helcom.html#rename_cols',
                                                                                  'marisco/handlers/helcom.py')},
            'marisco.handlers.maris_dump': { 'marisco.handlers.maris_dump.DropNAColumnsCB': ( 'handlers/maris_dump.html#dropnacolumnscb',
                                                                                              'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.DropNAColumnsCB.__call__': ( 'handlers/maris_dump.html#__call__',
                                                                                                       'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.DropNAColumnsCB.__init__': ( 'handlers/maris_dump.html#__init__',
                                                                                                       'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.DropNAColumnsCB.dropMarisNA': ( 'handlers/maris_dump.html#dropmarisna',
                                                                                                          'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.DropNAColumnsCB.isMarisNA': ( 'handlers/maris_dump.html#ismarisna',
                                                                                                        'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.ParseTimeCB': ( 'handlers/maris_dump.html#parsetimecb',
                                                                                          'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.ParseTimeCB.__call__': ( 'handlers/maris_dump.html#__call__',
                                                                                                   'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.RemapRdnNameCB': ( 'handlers/maris_dump.html#remaprdnnamecb',
                                                                                             'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.RemapRdnNameCB.__call__': ( 'handlers/maris_dump.html#__call__',
                                                                                                      'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.RemapRdnNameCB.__init__': ( 'handlers/maris_dump.html#__init__',
                                                                                                      'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.RenameColumnCB': ( 'handlers/maris_dump.html#renamecolumncb',
                                                                                             'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.RenameColumnCB.__call__': ( 'handlers/maris_dump.html#__call__',
                                                                                                      'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.RenameColumnCB.__init__': ( 'handlers/maris_dump.html#__init__',
                                                                                                      'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.ReshapeLongToWide': ( 'handlers/maris_dump.html#reshapelongtowide',
                                                                                                'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.ReshapeLongToWide.__call__': ( 'handlers/maris_dump.html#__call__',
                                                                                                         'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.ReshapeLongToWide.__init__': ( 'handlers/maris_dump.html#__init__',
                                                                                                         'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.ReshapeLongToWide.pivot': ( 'handlers/maris_dump.html#pivot',
                                                                                                      'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.ReshapeLongToWide.renamed_cols': ( 'handlers/maris_dump.html#renamed_cols',
                                                                                                             'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.SanitizeDetectionLimitCB': ( 'handlers/maris_dump.html#sanitizedetectionlimitcb',
                                                                                                       'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.SanitizeDetectionLimitCB.__call__': ( 'handlers/maris_dump.html#__call__',
                                                                                                                'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.SanitizeDetectionLimitCB.__init__': ( 'handlers/maris_dump.html#__init__',
                                                                                                                'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.encode': ( 'handlers/maris_dump.html#encode',
                                                                                     'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.enums_xtra': ( 'handlers/maris_dump.html#enums_xtra',
                                                                                         'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.get_attrs': ( 'handlers/maris_dump.html#get_attrs',
                                                                                        'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.get_dl_lut': ( 'handlers/maris_dump.html#get_dl_lut',
                                                                                         'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.get_fname': ( 'handlers/maris_dump.html#get_fname',
                                                                                        'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.get_varnames_lut': ( 'handlers/maris_dump.html#get_varnames_lut',
                                                                                               'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.get_zotero_key': ( 'handlers/maris_dump.html#get_zotero_key',
                                                                                             'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.load_data': ( 'handlers/maris_dump.html#load_data',
                                                                                        'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.load_dump': ( 'handlers/maris_dump.html#load_dump',
                                                                                        'marisco/handlers/maris_dump.py'),
                                             'marisco.handlers.maris_dump.renaming_rules': ( 'handlers/maris_dump.html#renaming_rules',
                                                                                             'marisco/handlers/maris_dump.py')},
            'marisco.handlers.tepco': { 'marisco.handlers.tepco.FixMissingValuesCB': ( 'handlers/tepco.html#fixmissingvaluescb',
                                                                                       'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.FixMissingValuesCB.__call__': ( 'handlers/tepco.html#__call__',
                                                                                                'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.FixRangeValueStringCB': ( 'handlers/tepco.html#fixrangevaluestringcb',
                                                                                          'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.FixRangeValueStringCB.__call__': ( 'handlers/tepco.html#__call__',
                                                                                                   'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.FixRangeValueStringCB._extract_and_calculate_mean': ( 'handlers/tepco.html#_extract_and_calculate_mean',
                                                                                                                      'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.FixRangeValueStringCB._transform_if_range': ( 'handlers/tepco.html#_transform_if_range',
                                                                                                              'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.ParseTimeCB': ( 'handlers/tepco.html#parsetimecb',
                                                                                'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.ParseTimeCB.__call__': ( 'handlers/tepco.html#__call__',
                                                                                         'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.ParseTimeCB.__init__': ( 'handlers/tepco.html#__init__',
                                                                                         'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.RemapRdnNameCB': ( 'handlers/tepco.html#remaprdnnamecb',
                                                                                   'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.RemapRdnNameCB.__call__': ( 'handlers/tepco.html#__call__',
                                                                                            'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.RemapRdnNameCB.__init__': ( 'handlers/tepco.html#__init__',
                                                                                            'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.RemoveJapanaseCharCB': ( 'handlers/tepco.html#removejapanasecharcb',
                                                                                         'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.RemoveJapanaseCharCB.__call__': ( 'handlers/tepco.html#__call__',
                                                                                                  'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.RemoveJapanaseCharCB._transform_if_about': ( 'handlers/tepco.html#_transform_if_about',
                                                                                                             'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.RenameColumnCB': ( 'handlers/tepco.html#renamecolumncb',
                                                                                   'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.RenameColumnCB.__call__': ( 'handlers/tepco.html#__call__',
                                                                                            'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.RenameColumnCB.__init__': ( 'handlers/tepco.html#__init__',
                                                                                            'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.SelectColumnsCB': ( 'handlers/tepco.html#selectcolumnscb',
                                                                                    'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.SelectColumnsCB.__call__': ( 'handlers/tepco.html#__call__',
                                                                                             'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.SelectColumnsCB.__init__': ( 'handlers/tepco.html#__init__',
                                                                                             'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.get_attrs': ('handlers/tepco.html#get_attrs', 'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.load_data': ('handlers/tepco.html#load_data', 'marisco/handlers/tepco.py'),
                                        'marisco.handlers.tepco.time_parser': ( 'handlers/tepco.html#time_parser',
                                                                                'marisco/handlers/tepco.py')},
            'marisco.inout': { 'marisco.inout.read_toml': ('api/inout.html#read_toml', 'marisco/inout.py'),
                               'marisco.inout.write_toml': ('api/inout.html#write_toml', 'marisco/inout.py')},
            'marisco.metadata': { 'marisco.metadata.BboxCB': ('api/metadata.html#bboxcb', 'marisco/metadata.py'),
                                  'marisco.metadata.BboxCB.__call__': ('api/metadata.html#__call__', 'marisco/metadata.py'),
                                  'marisco.metadata.DepthRangeCB': ('api/metadata.html#depthrangecb', 'marisco/metadata.py'),
                                  'marisco.metadata.DepthRangeCB.__call__': ('api/metadata.html#__call__', 'marisco/metadata.py'),
                                  'marisco.metadata.DepthRangeCB.__init__': ('api/metadata.html#__init__', 'marisco/metadata.py'),
                                  'marisco.metadata.GlobAttrsFeeder': ('api/metadata.html#globattrsfeeder', 'marisco/metadata.py'),
                                  'marisco.metadata.GlobAttrsFeeder.__call__': ('api/metadata.html#__call__', 'marisco/metadata.py'),
                                  'marisco.metadata.GlobAttrsFeeder.__init__': ('api/metadata.html#__init__', 'marisco/metadata.py'),
                                  'marisco.metadata.GlobAttrsFeeder.callback': ('api/metadata.html#callback', 'marisco/metadata.py'),
                                  'marisco.metadata.KeyValuePairCB': ('api/metadata.html#keyvaluepaircb', 'marisco/metadata.py'),
                                  'marisco.metadata.KeyValuePairCB.__call__': ('api/metadata.html#__call__', 'marisco/metadata.py'),
                                  'marisco.metadata.KeyValuePairCB.__init__': ('api/metadata.html#__init__', 'marisco/metadata.py'),
                                  'marisco.metadata.TimeRangeCB': ('api/metadata.html#timerangecb', 'marisco/metadata.py'),
                                  'marisco.metadata.TimeRangeCB.__call__': ('api/metadata.html#__call__', 'marisco/metadata.py'),
                                  'marisco.metadata.TimeRangeCB.__init__': ('api/metadata.html#__init__', 'marisco/metadata.py'),
                                  'marisco.metadata.ZoteroCB': ('api/metadata.html#zoterocb', 'marisco/metadata.py'),
                                  'marisco.metadata.ZoteroCB.__call__': ('api/metadata.html#__call__', 'marisco/metadata.py'),
                                  'marisco.metadata.ZoteroCB.__init__': ('api/metadata.html#__init__', 'marisco/metadata.py'),
                                  'marisco.metadata.ZoteroItem': ('api/metadata.html#zoteroitem', 'marisco/metadata.py'),
                                  'marisco.metadata.ZoteroItem.__init__': ('api/metadata.html#__init__', 'marisco/metadata.py'),
                                  'marisco.metadata.ZoteroItem.__repr__': ('api/metadata.html#__repr__', 'marisco/metadata.py'),
                                  'marisco.metadata.ZoteroItem.creator_name': ('api/metadata.html#creator_name', 'marisco/metadata.py'),
                                  'marisco.metadata.ZoteroItem.exist': ('api/metadata.html#exist', 'marisco/metadata.py'),
                                  'marisco.metadata.ZoteroItem.getItem': ('api/metadata.html#getitem', 'marisco/metadata.py'),
                                  'marisco.metadata.ZoteroItem.summary': ('api/metadata.html#summary', 'marisco/metadata.py'),
                                  'marisco.metadata.ZoteroItem.title': ('api/metadata.html#title', 'marisco/metadata.py')},
            'marisco.nc_template': { 'marisco.nc_template.NCTemplater': ('api/nc_template.html#nctemplater', 'marisco/nc_template.py'),
                                     'marisco.nc_template.NCTemplater.__init__': ( 'api/nc_template.html#__init__',
                                                                                   'marisco/nc_template.py'),
                                     'marisco.nc_template.NCTemplater.create_analyte_variables': ( 'api/nc_template.html#nctemplater.create_analyte_variables',
                                                                                                   'marisco/nc_template.py'),
                                     'marisco.nc_template.NCTemplater.create_default_variables': ( 'api/nc_template.html#nctemplater.create_default_variables',
                                                                                                   'marisco/nc_template.py'),
                                     'marisco.nc_template.NCTemplater.create_enum_types': ( 'api/nc_template.html#nctemplater.create_enum_types',
                                                                                            'marisco/nc_template.py'),
                                     'marisco.nc_template.NCTemplater.create_group_specific_variables': ( 'api/nc_template.html#nctemplater.create_group_specific_variables',
                                                                                                          'marisco/nc_template.py'),
                                     'marisco.nc_template.NCTemplater.create_groups': ( 'api/nc_template.html#nctemplater.create_groups',
                                                                                        'marisco/nc_template.py'),
                                     'marisco.nc_template.NCTemplater.create_variable': ( 'api/nc_template.html#nctemplater.create_variable',
                                                                                          'marisco/nc_template.py'),
                                     'marisco.nc_template.NCTemplater.create_variables': ( 'api/nc_template.html#nctemplater.create_variables',
                                                                                           'marisco/nc_template.py'),
                                     'marisco.nc_template.NCTemplater.derive': ( 'api/nc_template.html#nctemplater.derive',
                                                                                 'marisco/nc_template.py'),
                                     'marisco.nc_template.NCTemplater.generate': ( 'api/nc_template.html#nctemplater.generate',
                                                                                   'marisco/nc_template.py'),
                                     'marisco.nc_template.NCTemplater.nuclide_vars': ( 'api/nc_template.html#nctemplater.nuclide_vars',
                                                                                       'marisco/nc_template.py')},
            'marisco.serializers': { 'marisco.serializers.NetCDFEncoder': ('api/serializers.html#netcdfencoder', 'marisco/serializers.py'),
                                     'marisco.serializers.NetCDFEncoder.__init__': ( 'api/serializers.html#__init__',
                                                                                     'marisco/serializers.py'),
                                     'marisco.serializers.NetCDFEncoder._create_and_copy_variable': ( 'api/serializers.html#netcdfencoder._create_and_copy_variable',
                                                                                                      'marisco/serializers.py'),
                                     'marisco.serializers.NetCDFEncoder.cast_verbose_rf': ( 'api/serializers.html#netcdfencoder.cast_verbose_rf',
                                                                                            'marisco/serializers.py'),
                                     'marisco.serializers.NetCDFEncoder.copy_dimensions': ( 'api/serializers.html#netcdfencoder.copy_dimensions',
                                                                                            'marisco/serializers.py'),
                                     'marisco.serializers.NetCDFEncoder.copy_enum_type': ( 'api/serializers.html#netcdfencoder.copy_enum_type',
                                                                                           'marisco/serializers.py'),
                                     'marisco.serializers.NetCDFEncoder.copy_global_attributes': ( 'api/serializers.html#netcdfencoder.copy_global_attributes',
                                                                                                   'marisco/serializers.py'),
                                     'marisco.serializers.NetCDFEncoder.copy_variable': ( 'api/serializers.html#netcdfencoder.copy_variable',
                                                                                          'marisco/serializers.py'),
                                     'marisco.serializers.NetCDFEncoder.copy_variable_attributes': ( 'api/serializers.html#netcdfencoder.copy_variable_attributes',
                                                                                                     'marisco/serializers.py'),
                                     'marisco.serializers.NetCDFEncoder.copy_variables': ( 'api/serializers.html#netcdfencoder.copy_variables',
                                                                                           'marisco/serializers.py'),
                                     'marisco.serializers.NetCDFEncoder.encode': ( 'api/serializers.html#netcdfencoder.encode',
                                                                                   'marisco/serializers.py'),
                                     'marisco.serializers.NetCDFEncoder.process_group': ( 'api/serializers.html#netcdfencoder.process_group',
                                                                                          'marisco/serializers.py'),
                                     'marisco.serializers.NetCDFEncoder.process_groups': ( 'api/serializers.html#netcdfencoder.process_groups',
                                                                                           'marisco/serializers.py')},
            'marisco.utils': { 'marisco.utils.Callback': ('api/utils.html#callback', 'marisco/utils.py'),
                               'marisco.utils.Transformer': ('api/utils.html#transformer', 'marisco/utils.py'),
                               'marisco.utils.Transformer.__call__': ('api/utils.html#__call__', 'marisco/utils.py'),
                               'marisco.utils.Transformer.__init__': ('api/utils.html#__init__', 'marisco/utils.py'),
                               'marisco.utils.Transformer.callback': ('api/utils.html#callback', 'marisco/utils.py'),
                               'marisco.utils.download_file': ('api/utils.html#download_file', 'marisco/utils.py'),
                               'marisco.utils.download_files_in_folder': ('api/utils.html#download_files_in_folder', 'marisco/utils.py'),
                               'marisco.utils.get_bbox': ('api/utils.html#get_bbox', 'marisco/utils.py'),
                               'marisco.utils.has_valid_varname': ('api/utils.html#has_valid_varname', 'marisco/utils.py'),
                               'marisco.utils.match_maris_sediment': ('api/utils.html#match_maris_sediment', 'marisco/utils.py'),
                               'marisco.utils.match_maris_species': ('api/utils.html#match_maris_species', 'marisco/utils.py'),
                               'marisco.utils.match_worms': ('api/utils.html#match_worms', 'marisco/utils.py'),
                               'marisco.utils.run_cbs': ('api/utils.html#run_cbs', 'marisco/utils.py')}}}